"""Tests for grizzly_cli.argparse.bashcompletion.types."""

from __future__ import annotations

from argparse import ArgumentTypeError
from os import sep
from typing import TYPE_CHECKING

import pytest
from grizzly_cli.argparse.bashcompletion.types import BashCompletionTypes

from test_cli.helpers import cwd, rm_rf

if TYPE_CHECKING:
    from _pytest.tmpdir import TempPathFactory


class TestBashCompletionTypes:
    class TestFile:
        def test___init__(self) -> None:
            assert BashCompletionTypes.File('*.txt').patterns == ['*.txt']
            assert BashCompletionTypes.File('*.txt', '*.json').patterns == ['*.txt', '*.json']
            assert BashCompletionTypes.File('*.txt', '*.json', '*.xml').patterns == ['*.txt', '*.json', '*.xml']

        def test___call__(self, tmp_path_factory: TempPathFactory) -> None:
            test_context = tmp_path_factory.mktemp('test_context')
            test_dir = test_context / 'test-dir'
            test_dir.mkdir()
            file = test_context / 'test.txt'
            file.touch()
            file.write_text('test.txt file')

            file = test_context / 'test.json'
            file.touch()
            file.write_text('{"value": "test.json file"}')

            file = test_context / 'test.xml'
            file.touch()
            file.write_text('<value>test.xml file</value>')

            try:
                with cwd(test_context):
                    impl = BashCompletionTypes.File('*.txt')

                    with pytest.raises(ArgumentTypeError) as ate:
                        impl('non-existing-directory/')
                    assert 'non-existing-directory/ does not exist' in str(ate)

                    with pytest.raises(ArgumentTypeError) as ate:
                        impl('test-dir/')
                    assert 'test-dir/ is not a file' in str(ate)

                    with pytest.raises(ArgumentTypeError) as ate:
                        impl('test.xml')
                    assert 'test.xml does not match *.txt' in str(ate)

                    assert impl('test.txt') == 'test.txt'

                    impl = BashCompletionTypes.File('*.txt', '*.json')

                    with pytest.raises(ArgumentTypeError) as ate:
                        impl('test.xml')
                    assert 'test.xml does not match *.txt' in str(ate)

                    assert impl('test.txt') == 'test.txt'
                    assert impl('test.json') == 'test.json'
            finally:
                rm_rf(test_context)

        def test_list_files(self, tmp_path_factory: TempPathFactory) -> None:
            test_context = tmp_path_factory.mktemp('test_context')
            (test_context / 'test.txt').write_text('test.txt file')
            (test_context / 'test.json').write_text('{"value": "test.json file"}')
            (test_context / 'test.xml').write_text('<value>test.xml file</value>')

            test_dir = test_context / 'test-dir'
            test_dir.mkdir()
            (test_dir / 'test.txt').write_text('sub test.txt file')

            hidden_dir = test_context / '.hidden'
            hidden_dir.mkdir()
            (hidden_dir / 'hidden.txt').write_text('hidden.txt file')
            try:
                with cwd(test_context):
                    impl = BashCompletionTypes.File('*.txt')
                    assert impl.list_files(None) == {
                        'test.txt': 'file',
                        'test-dir': 'dir',
                    }
                    assert impl.list_files('te') == {
                        'test.txt': 'file',
                        'test-dir': 'dir',
                    }

                    assert impl.list_files('test-') == {
                        'test-dir': 'dir',
                    }

                    assert impl.list_files(f'test-dir{sep}') == {
                        f'test-dir{sep}test.txt': 'file',
                    }

                    impl = BashCompletionTypes.File('*.txt', '*.json', '*.xml')
                    assert impl.list_files('te') == {
                        'test.txt': 'file',
                        'test.json': 'file',
                        'test.xml': 'file',
                        'test-dir': 'dir',
                    }
            finally:
                rm_rf(test_context)



"""Tests for grizzly_cli.keyvault."""

from __future__ import annotations

from typing import TYPE_CHECKING

from grizzly_cli.keyvault import (
    COMMON_FALSE_POSITIVES,
    KEYWORDS,
    KeyvaultSecretHolder,
    _build_key_name,
    _determine_environment,
    _dict_to_yaml,
    _extract_metadata,
    _keyvault_normalize,
    _should_export,
    encode_file,
    encode_mq_certificate,
)

from test_cli.helpers import SOME, rm_rf

if TYPE_CHECKING:  # pragma: no cover
    from _pytest.tmpdir import TempPathFactory
    from pytest_mock.plugin import MockerFixture


def test__keyvault_normalize() -> None:
    assert _keyvault_normalize('fo0b4R') == 'fo0b4R'
    assert _keyvault_normalize('hello.world!') == 'hello-world-'


def test_encode_mq_certificate(mocker: MockerFixture, tmp_path_factory: TempPathFactory) -> None:
    glob_mock = mocker.patch('pathlib.Path.glob', return_value=None)
    root = tmp_path_factory.mktemp('test_context')

    try:
        # create "certificate" files
        kdb_file = root / 'foobar.kdb'
        kdb_file.write_bytes(b'A' * 25501)
        sth_file = root / 'foobar.sth'
        sth_file.write_bytes(b'B' * 255)

        glob_mock.return_value = iter([kdb_file, sth_file])

        encoded_mq_certificates = encode_mq_certificate(root, 'test', 'grizzly--test--mq-keyfile', 'foobar')
        assert encoded_mq_certificates == [
            SOME(KeyvaultSecretHolder, name='grizzly--test--foobar-kdb--0', content_type='file:foobar.kdb,chunk:0,chunks:2,noconf'),
            SOME(KeyvaultSecretHolder, name='grizzly--test--foobar-kdb--1', content_type='file:foobar.kdb,chunk:1,chunks:2,noconf'),
            SOME(KeyvaultSecretHolder, name='grizzly--test--foobar-kdb', content_type='files,noconf', value='grizzly--test--foobar-kdb--0,grizzly--test--foobar-kdb--1'),
            SOME(KeyvaultSecretHolder, name='grizzly--test--foobar-sth', content_type='file:foobar.sth,noconf'),
            SOME(KeyvaultSecretHolder, name='grizzly--test--mq-keyfile', content_type='files', value='grizzly--test--foobar-kdb,grizzly--test--foobar-sth'),
        ]
    finally:
        rm_rf(root)


def test_encode_file(tmp_path_factory: TempPathFactory) -> None:
    root = tmp_path_factory.mktemp('test_context')

    try:
        file = root / 'test.txt'
        file.write_bytes(b'C' * 512)

        keyvault_file = encode_file('grizzly--test--file', file.as_posix(), no_conf=False)
        assert keyvault_file == [
            SOME(KeyvaultSecretHolder, name='grizzly--test--file', content_type='file:test.txt'),
        ]
    finally:
        rm_rf(root)


def test__should_export() -> None:
    assert not _should_export('keyvault', 'test.vault.azure.net')
    assert not _should_export('hello.world', 'foobar')

    for keyword in KEYWORDS:
        assert _should_export(f'user.{keyword}', 'foobar')
        assert not _should_export(f'user.{keyword}.description', 'foobar')
        assert _should_export('user.description', f'hello{keyword.upper()}')

        for common_false_positive in COMMON_FALSE_POSITIVES:
            assert not _should_export('user.description', common_false_positive)


def test__determine_environment() -> None:
    assert _determine_environment([], 'test', 'foo.bar') == 'test'
    assert _determine_environment(['foo'], 'test', 'foo.bar') == 'global'


def test__build_key_name() -> None:
    assert _build_key_name('test', 'foo.bar') == 'grizzly--test--foo-bar'
    assert _build_key_name('global', 'hello.world.foo.bar') == 'grizzly--global--hello-world-foo-bar'


def test__dict_to_yaml(tmp_path_factory: TempPathFactory) -> None:
    root = tmp_path_factory.mktemp('test_context')

    try:
        file = root / 'test.yaml'
        file.write_text('lorem ipsum')

        content: dict = {
            'foo': {
                'bar': 'hello world',
            },
            'hello': 'world',
            'test': {
                'struct': {
                    'with': [
                        'value1',
                        'value2',
                    ],
                },
            },
        }

        _dict_to_yaml(file, content, indentation=8)

        assert (
            file.read_text()
            == """foo:
        bar: hello world
hello: world
test:
        struct:
                with:
                        - value1
                        - value2
"""
        )
        _dict_to_yaml(file, content, indentation=2)

        assert (
            file.read_text()
            == """foo:
  bar: hello world
hello: world
test:
  struct:
    with:
      - value1
      - value2
"""
        )
    finally:
        rm_rf(root)


def test__extract_metadata(tmp_path_factory: TempPathFactory) -> None:
    root = tmp_path_factory.mktemp('test_context')

    try:
        env_file = root / 'test.yaml'
        env = {
            'configuration': {
                'keyvault': 'https://test.vault.azure.net',
                'env': 'test',
                'foo': {
                    'bar': 'hello world',
                },
            },
        }

        _dict_to_yaml(env_file, env, indentation=2)

        assert _extract_metadata(env_file.as_posix()) == (
            'test',
            'https://test.vault.azure.net',
            {
                'keyvault': 'https://test.vault.azure.net',
                'env': 'test',
                'foo.bar': 'hello world',
            },
        )

        env = {
            'configuration': {
                'foo': {
                    'bar': 'hello world',
                },
            },
        }

        _dict_to_yaml(env_file, env, indentation=2)

        assert _extract_metadata(env_file.as_posix()) == (
            'test',
            None,
            {
                'foo.bar': 'hello world',
            },
        )
    finally:
        rm_rf(root)



"""Functionality for `grizzly-cli dist build ...`."""

from __future__ import annotations

import os
import sys
from argparse import SUPPRESS
from argparse import Namespace as Arguments
from getpass import getuser
from pathlib import Path
from socket import gaierror, gethostbyname
from typing import TYPE_CHECKING

from grizzly_cli import EXECUTION_CONTEXT, PROJECT_NAME, STATIC_CONTEXT
from grizzly_cli.utils import get_dependency_versions, requirements, run_command

if TYPE_CHECKING:  # pragma: no cover
    from grizzly_cli.argparse import ArgumentSubParser


def create_parser(sub_parser: ArgumentSubParser) -> None:
    # grizzly-cli dist build ...
    build_parser = sub_parser.add_parser(
        'build',
        description=(
            'build grizzly compose project container image before running test. if worker nodes runs on different physical '
            'computers, it is mandatory to build the images before hand and push to a registry.'
            '\n\n'
            'if image includes IBM MQ native dependencies, the build time increases due to download times. it is possible '
            'to self-host the archive and override the download host with environment variable `IBM_MQ_LIB_HOST`.'
        ),
    )
    build_parser.add_argument(
        '--no-cache',
        action='store_true',
        required=False,
        help='build container image with out cache (full build)',
    )
    build_parser.add_argument(
        '--registry',
        type=str,
        default=None,
        required=False,
        help='push built image to this registry, if the registry has authentication you need to login first',
    )
    build_parser.add_argument(
        '--no-progress',
        action='store_true',
        default=False,
        required=False,
        help='do not show a progress spinner while building',
    )
    build_parser.add_argument(
        '--verbose',
        action='store_true',
        default=False,
        required=False,
        help='show more information',
    )
    # <!-- used during development, hide from help
    build_parser.add_argument(
        '--local-install',
        nargs='?',
        const=True,
        default=False,
        help=SUPPRESS,
    )
    # used during development, hide from help -->

    if build_parser.prog != 'grizzly-cli dist build':  # pragma: no cover
        build_parser.prog = 'grizzly-cli dist build'


def getuid() -> int:
    if sys.platform == 'win32':
        return 1000
    else:  # noqa: RET505
        return os.getuid()


def getgid() -> int:
    if sys.platform == 'win32':
        return 1000
    else:  # noqa: RET505
        return os.getuid()


def _create_build_command(args: Arguments, containerfile: str, tag: str, context: str) -> list[str]:
    local_install = getattr(args, 'local_install', False)

    install_type = 'local' if local_install else 'remote'

    (_, grizzly_extras), _ = get_dependency_versions(local_install=local_install)

    grizzly_extra = 'mq' if grizzly_extras is not None and 'mq' in grizzly_extras else 'base'

    extra_args: list[str] = []

    ibm_mq_lib_host = os.environ.get('IBM_MQ_LIB_HOST', None)
    if ibm_mq_lib_host is not None:
        extra_args += ['--build-arg', f'IBM_MQ_LIB_HOST={ibm_mq_lib_host}']

        if 'host.docker.internal' in ibm_mq_lib_host:
            try:
                host_docker_internal = gethostbyname('host.docker.internal')
            except gaierror:
                host_docker_internal = 'host-gateway'

            extra_args += ['--add-host', f'host.docker.internal:{host_docker_internal}']

    ibm_mq_lib = os.environ.get('IBM_MQ_LIB', None)
    if ibm_mq_lib is not None:
        extra_args += ['--build-arg', f'IBM_MQ_LIB={ibm_mq_lib}']

    return [
        args.container_system,
        'image',
        'build',
        '--ssh',
        'default',
        '--build-arg',
        f'GRIZZLY_EXTRA={grizzly_extra}',
        '--build-arg',
        f'GRIZZLY_INSTALL_TYPE={install_type}',
        '--build-arg',
        f'GRIZZLY_UID={getuid()}',
        '--build-arg',
        f'GRIZZLY_GID={getgid()}',
        *extra_args,
        '-f',
        containerfile,
        '-t',
        tag,
        context,
    ]


@requirements(EXECUTION_CONTEXT)
def build(args: Arguments) -> int:
    tag = getuser()

    image_name = f'{PROJECT_NAME}:{tag}' if args.project_name is None else f'{args.project_name}:{tag}'

    build_command = _create_build_command(
        args,
        Path.joinpath(Path(STATIC_CONTEXT), 'Containerfile').as_posix(),
        image_name,
        EXECUTION_CONTEXT,
    )

    if args.force_build:
        build_command.append('--no-cache')

    # make sure buildkit is used
    build_env = os.environ.copy()
    if args.container_system == 'docker':
        build_env['DOCKER_BUILDKIT'] = '1'

    spinner = 'building' if not getattr(args, 'no_progress', False) else None

    result = run_command(build_command, env=build_env, spinner=spinner, verbose=args.verbose)

    if result.return_code == 0:
        print(f'\nbuilt image {image_name}')

    if getattr(args, 'registry', None) is None or result.return_code != 0:
        return result.return_code

    tag_command = [
        f'{args.container_system}',
        'image',
        'tag',
        image_name,
        f'{args.registry}{image_name}',
    ]

    result = run_command(tag_command, env=build_env, verbose=args.verbose)

    if result.return_code != 0:
        print(f'\n!! failed to tag image {image_name} -> {args.registry}{image_name}')
        return result.return_code

    print(f'tagged image {image_name} -> {args.registry}{image_name}')

    push_command = [
        f'{args.container_system}',
        'image',
        'push',
        f'{args.registry}{image_name}',
    ]

    spinner = 'pushing' if not args.no_progress else None

    result = run_command(push_command, env=build_env, spinner=spinner, verbose=args.verbose)

    if result.return_code != 0:
        print(f'\n!! failed to push image {args.registry}{image_name}')
    else:
        print(f'pushed image {args.registry}{image_name}')

    return result.return_code



"""Tests for grizzly-cli auth."""

from __future__ import annotations

import re
from contextlib import contextmanager, suppress
from os import environ
from typing import TYPE_CHECKING

import pytest

from test_cli.helpers import rm_rf, run_command

if TYPE_CHECKING:
    from collections.abc import Generator

    from _pytest.tmpdir import TempPathFactory


@contextmanager
def auth_via(tmp_path_factory: TempPathFactory, method: str) -> Generator[tuple[str | None, str | None], None, None]:
    secret = 'asdfasdf'  # noqa: S105
    test_context = tmp_path_factory.mktemp('test_context')
    argument: str | None = None
    stdin: str | None = None

    if method == 'env':
        environ['OTP_SECRET'] = secret
        argument = None
    elif method == 'stdin':
        argument = '-'
        stdin = secret
    elif method == 'file':
        file = test_context / 'secret.txt'
        file.write_text(f'{secret}\n')
        argument = str(file)

    try:
        yield (argument, stdin)
    finally:
        if method == 'env':
            with suppress(KeyError):
                del environ['OTP_SECRET']

        rm_rf(test_context)


@pytest.mark.parametrize('method', ['env', 'file', 'stdin'])
def test_e2e_auth(tmp_path_factory: TempPathFactory, method: str) -> None:
    with auth_via(tmp_path_factory, method) as context:
        argument, stdin = context
        command = ['grizzly-cli', 'auth']
        if argument is not None:
            command.append(argument)

        rc, output = run_command(command, stdin=stdin)

        assert rc == 0

        result = ''.join(output).strip()

        assert re.match(r'^[0-9]{6}$', result)



"""Tests for grizzly_cli:main."""

from __future__ import annotations

import sys
from argparse import ArgumentParser as CoreArgumentParser
from argparse import Namespace
from hashlib import sha1
from os import environ
from pathlib import Path
from typing import TYPE_CHECKING, cast

import pytest
from grizzly_cli.__main__ import _create_parser, _inject_additional_arguments_from_metadata, _parse_arguments, main

from test_cli.helpers import SOME, cwd, rm_rf

if TYPE_CHECKING:
    from _pytest.capture import CaptureFixture
    from _pytest.tmpdir import TempPathFactory
    from pytest_mock import MockerFixture


def test__create_parser() -> None:  # noqa: PLR0915
    parser = _create_parser()

    assert parser.prog == 'grizzly-cli'
    assert parser.description is not None
    assert 'pip install grizzly-loadtester-cli' in parser.description
    assert 'eval "$(grizzly-cli --bash-completion)"' in parser.description
    assert parser._subparsers is not None
    assert len(parser._subparsers._group_actions) == 1
    assert sorted([option_string for action in parser._actions for option_string in action.option_strings]) == sorted(
        [
            '-h',
            '--help',
            '--version',
            '--md-help',
            '--bash-completion',
        ]
    )
    assert sorted([action.dest for action in parser._actions if len(action.option_strings) == 0]) == ['command']
    subparser = parser._subparsers._group_actions[0]
    assert subparser is not None
    assert subparser.choices is not None
    assert len(cast('dict[str, CoreArgumentParser | None]', subparser.choices).keys()) == 5

    init_parser = cast('dict[str, CoreArgumentParser | None]', subparser.choices).get('init', None)
    assert init_parser is not None
    assert init_parser._subparsers is None
    assert getattr(init_parser, 'prog', None) == 'grizzly-cli init'
    assert sorted([option_string for action in init_parser._actions for option_string in action.option_strings]) == sorted(
        [
            '-h',
            '--help',
            '-y',
            '--yes',
            '--with-mq',
            '--grizzly-version',
        ]
    )

    auth_parser = cast('dict[str, CoreArgumentParser | None]', subparser.choices).get('auth', None)
    assert auth_parser is not None
    assert auth_parser._subparsers is None
    assert getattr(auth_parser, 'prog', None) == 'grizzly-cli auth'
    assert sorted([option_string for action in auth_parser._actions for option_string in action.option_strings]) == sorted(
        [
            '-h',
            '--help',
        ]
    )

    keyvault_parser = cast('dict[str, CoreArgumentParser | None]', subparser.choices).get('keyvault', None)
    assert keyvault_parser is not None
    print(keyvault_parser._subparsers)
    assert keyvault_parser._subparsers is not None
    assert getattr(keyvault_parser, 'prog', None) == 'grizzly-cli keyvault'
    assert sorted([option_string for action in keyvault_parser._actions for option_string in action.option_strings]) == sorted(
        [
            '--file',
            '-f',
            '-h',
            '--help',
            '--vault-name',
        ]
    )

    local_parser = cast('dict[str, CoreArgumentParser | None]', subparser.choices).get('local', None)
    assert local_parser is not None
    assert local_parser._subparsers is not None
    assert getattr(local_parser, 'prog', None) == 'grizzly-cli local'
    assert sorted([option_string for action in local_parser._actions for option_string in action.option_strings]) == sorted(
        [
            '-h',
            '--help',
        ]
    )
    assert len(local_parser._subparsers._group_actions) == 1
    local_subparser = local_parser._subparsers._group_actions[0]
    assert local_subparser is not None
    assert local_subparser.choices is not None
    assert list(cast('dict[str, CoreArgumentParser | None]', local_subparser.choices).keys()) == ['run']

    dist_parser = cast('dict[str, CoreArgumentParser | None]', subparser.choices).get('dist', None)
    assert dist_parser is not None
    assert dist_parser._subparsers is not None
    assert getattr(dist_parser, 'prog', None) == 'grizzly-cli dist'
    assert sorted([option_string for action in dist_parser._actions for option_string in action.option_strings]) == sorted(
        [
            '-h',
            '--help',
            '--force-build',
            '--build',
            '--validate-config',
            '--workers',
            '--id',
            '--limit-nofile',
            '--container-system',
            '--health-timeout',
            '--health-retries',
            '--health-interval',
            '--project-name',
            '--registry',
            '--tty',
            '--wait-for-worker',
        ]
    )
    assert sorted([action.dest for action in dist_parser._actions if len(action.option_strings) == 0]) == ['subcommand']
    assert len(dist_parser._subparsers._group_actions) == 1
    dist_subparser = dist_parser._subparsers._group_actions[0]
    assert dist_subparser is not None
    assert dist_subparser.choices is not None
    assert list(cast('dict[str, CoreArgumentParser | None]', dist_subparser.choices).keys()) == ['build', 'clean', 'run']

    dist_build_parser = cast('dict[str, CoreArgumentParser | None]', dist_subparser.choices).get('build', None)
    assert dist_build_parser is not None
    assert dist_build_parser._subparsers is None
    assert getattr(dist_build_parser, 'prog', None) == 'grizzly-cli dist build'
    assert sorted([option_string for action in dist_build_parser._actions for option_string in action.option_strings]) == sorted(
        [
            '-h',
            '--help',
            '--local-install',
            '--no-cache',
            '--registry',
            '--no-progress',
            '--verbose',
        ]
    )

    dist_clean_parser = cast('dict[str, CoreArgumentParser | None]', dist_subparser.choices).get('clean', None)
    assert dist_clean_parser is not None
    assert dist_clean_parser._subparsers is None
    assert getattr(dist_clean_parser, 'prog', None) == 'grizzly-cli dist clean'
    assert sorted([option_string for action in dist_clean_parser._actions for option_string in action.option_strings]) == sorted(
        [
            '-h',
            '--help',
            '--no-images',
            '--no-networks',
        ]
    )

    # grizzly-cli ... run
    for tested_parser, parent in [(local_parser, 'local'), (dist_parser, 'dist')]:
        assert tested_parser._subparsers is not None
        assert len(tested_parser._subparsers._group_actions) == 1
        subparser = tested_parser._subparsers._group_actions[0]
        run_parser = cast('dict[str, CoreArgumentParser | None]', subparser.choices).get('run', None)
        assert run_parser is not None
        assert getattr(run_parser, 'prog', None) == f'grizzly-cli {parent} run'
        assert sorted([option_string for action in run_parser._actions for option_string in action.option_strings]) == sorted(
            [
                '-h',
                '--help',
                '--verbose',
                '-T',
                '--testdata-variable',
                '-y',
                '--yes',
                '-e',
                '--environment-file',
                '--csv-prefix',
                '--csv-interval',
                '--csv-flush-interval',
                '-l',
                '--log-dir',
                '--log-file',
                '--dump',
                '--dry-run',
                '--profile',
            ]
        )
        assert sorted([action.dest for action in run_parser._actions if len(action.option_strings) == 0]) == ['file']


def test__parse_argument_version(capsys: CaptureFixture, mocker: MockerFixture, tmp_path_factory: TempPathFactory) -> None:  # noqa: PLR0915
    test_context = tmp_path_factory.mktemp('test_context')
    (test_context / 'test.feature').write_text('Feature:')

    import sys

    try:
        mocker.patch('grizzly_cli.EXECUTION_CONTEXT', test_context.as_posix())
        with cwd(test_context):
            sys.argv = ['grizzly-cli', '--version']

            expected_version = '0.0.0'
            expected_common_version = '1.2.3'

            mocker.patch('grizzly_cli.__main__.__version__', expected_version)
            mocker.patch('grizzly_cli.__main__.__common_version__', expected_common_version)

            with pytest.raises(SystemExit) as se:
                _parse_arguments()
            assert se.value.code == 0

            capture = capsys.readouterr()
            assert capture.err == ''
            assert capture.out == f'grizzly-cli {expected_version}\n└── grizzly-common {expected_common_version}\n'

            sys.argv = ['grizzly-cli', '--version', 'foo']

            with pytest.raises(SystemExit) as se:
                _parse_arguments()
            assert se.value.code == 2

            capture = capsys.readouterr()
            err = capture.err.split('\n')
            assert len(err) == 4
            assert err[0].startswith('usage: grizzly-cli')
            assert 'init,local,auth,dist,keyvault' in err[1]
            assert err[2] == ("grizzly-cli: error: argument --version: invalid choice: 'foo' (choose from 'all')") or (
                "grizzly-cli: error: argument --version: invalid choice: 'foo' (choose from all)"
            )
            assert err[3] == ''
            assert capture.out == ''

            requirements_file = test_context / 'requirements.txt'
            requirements_file.write_text('grizzly-loadtester==1.5.3\n')

            sys.argv = ['grizzly-cli', '--version', 'all']

            with pytest.raises(SystemExit) as se:
                _parse_arguments()
            assert se.value.code == 0

            capture = capsys.readouterr()
            assert capture.err == ''
            assert capture.out == f'grizzly-cli {expected_version}\n├── grizzly-common {expected_common_version}\n└── grizzly 1.5.3\n    └── locust 2.2.1\n'

            requirements_file.write_text('grizzly-loadtester[mq]==1.5.3\n')

            sys.argv = ['grizzly-cli', '--version', 'all']

            with pytest.raises(SystemExit) as se:
                _parse_arguments()
            assert se.value.code == 0
            capture = capsys.readouterr()
            assert capture.err == ''
            assert capture.out == f'grizzly-cli {expected_version}\n├── grizzly-common {expected_common_version}\n└── grizzly 1.5.3 ── extras: mq\n    └── locust 2.2.1\n'

            requirements_file.unlink()
            requirements_file.write_text('grizzly-loadtester[mq,dev]==1.5.3\n')

            sys.argv = ['grizzly-cli', '--version', 'all']

            with pytest.raises(SystemExit) as se:
                _parse_arguments()
            assert se.value.code == 0
            capture = capsys.readouterr()
            assert capture.err == ''
            assert capture.out == (f'grizzly-cli {expected_version}\n├── grizzly-common {expected_common_version}\n└── grizzly 1.5.3 ── extras: mq, dev\n    └── locust 2.2.1\n')

            def mocked_mkdtemp(prefix: str | None = '') -> str:
                return Path.joinpath(test_context, f'{prefix}test').as_posix()

            mocker.patch('grizzly_cli.utils.mkdtemp', mocked_mkdtemp)
            mocker.patch('grizzly_cli.utils.subprocess.check_call', return_value=0)
            mocker.patch('grizzly_cli.utils.subprocess.check_output', return_value='main\n')

            repo = 'git+https://git@github.com/biometria-se/grizzly.git@main#egg=grizzly-loadtester'
            repo_suffix = sha1(repo.encode('utf-8')).hexdigest()  # noqa: S324
            repo_dir = test_context / 'grizzly-cli-test' / f'grizzly-loadtester_{repo_suffix}'
            repo_dir.mkdir(parents=True)
            (repo_dir / 'pyproject.toml').touch()
            (repo_dir / 'setup.cfg').write_text('name = grizzly-loadtester\nversion = 0.0.0\n')
            (repo_dir / 'requirements.txt').write_text('locust==2.8.4  \\ \n')

            requirements_file.unlink()
            requirements_file.write_text(f'{repo}\n')

            sys.argv = ['grizzly-cli', '--version', 'all']

            with pytest.raises(SystemExit) as se:
                _parse_arguments()
            assert se.value.code == 0

            capture = capsys.readouterr()
            assert capture.err == ''
            assert capture.out == f'grizzly-cli {expected_version}\n├── grizzly-common {expected_common_version}\n└── grizzly 0.0.0\n    └── locust 2.8.4\n'

            repo = 'git+https://git@github.com/biometria-se/grizzly.git@main#egg=grizzly-loadtester[mq,dev]'
            repo_suffix = sha1(repo.encode('utf-8')).hexdigest()  # noqa: S324
            repo_dir = test_context / 'grizzly-cli-test' / f'grizzly-loadtester__mq_dev___{repo_suffix}'
            repo_dir.mkdir(parents=True)
            (repo_dir / 'pyproject.toml').touch()
            (repo_dir / 'setup.cfg').write_text('name = grizzly-loadtester\nversion = 0.0.0\n')
            (repo_dir / 'requirements.txt').write_text('locust==2.8.4  \\ \n')

            requirements_file.unlink()
            requirements_file.write_text(f'{repo}\n')

            sys.argv = ['grizzly-cli', '--version', 'all']

            with pytest.raises(SystemExit) as se:
                _parse_arguments()
            assert se.value.code == 0

            capture = capsys.readouterr()
            assert capture.err == ''
            assert capture.out == f'grizzly-cli {expected_version}\n├── grizzly-common {expected_common_version}\n└── grizzly 0.0.0 ── extras: mq, dev\n    └── locust 2.8.4\n'

            requirements_file.unlink()
            requirements_file.write_text('grizzly-loadtester==1.5.3\n')

            sys.argv = ['grizzly-cli', '--version', 'all']
            expected_version = '2.5.0'
            expected_common_version = '2.4.0'
            mocker.patch('grizzly_cli.__main__.__version__', expected_version)
            mocker.patch('grizzly_cli.__main__.__common_version__', expected_common_version)

            with pytest.raises(SystemExit) as se:
                _parse_arguments()
            assert se.value.code == 0

            capture = capsys.readouterr()
            assert capture.err == ''
            assert capture.out == f'grizzly-cli {expected_version}\n├── grizzly-common {expected_common_version}\n└── grizzly 1.5.3\n    └── locust 2.2.1\n'
    finally:
        rm_rf(test_context)


def test__parse_argument_local(capsys: CaptureFixture, mocker: MockerFixture, tmp_path_factory: TempPathFactory) -> None:  # noqa: PLR0915
    test_context = tmp_path_factory.mktemp('test_context')
    (test_context / 'test.feature').write_text('Feature:')

    import sys

    try:
        mocker.patch('grizzly_cli.EXECUTION_CONTEXT', test_context.as_posix())
        with cwd(test_context):
            sys.argv = ['grizzly-cli', 'local']

            with pytest.raises(SystemExit) as se:
                _parse_arguments()
            assert se.value.code == 2

            capture = capsys.readouterr()
            assert capture.out == ''
            assert capture.err == 'grizzly-cli: error: no subcommand for local specified\n'

            sys.argv = ['grizzly-cli', 'local', 'run', 'test.feature']
            mocker.patch('grizzly_cli.__main__.which', side_effect=[None])

            with pytest.raises(SystemExit) as se:
                _parse_arguments()
            assert se.value.code == 2

            capture = capsys.readouterr()
            assert capture.out == ''
            assert capture.err == 'grizzly-cli: error: "behave" not found in PATH, needed when running local mode\n'

            # csv logging
            sys.argv = ['grizzly-cli', 'local', 'run', '--csv-interval', '20', 'test.feature']
            mocker.patch('grizzly_cli.__main__.which', side_effect=['behave'])

            with pytest.raises(SystemExit) as se:
                _parse_arguments()
            assert se.value.code == 2

            capture = capsys.readouterr()
            assert capture.out == ''
            assert capture.err == 'grizzly-cli: error: --csv-interval can only be used in combination with --csv-prefix\n'

            sys.argv = ['grizzly-cli', 'local', 'run', '--csv-prefix', '--csv-interval', '20', '--csv-flush-interval', '60', 'test.feature']
            mocker.patch('grizzly_cli.__main__.which', side_effect=['behave'])

            parsed_args = _parse_arguments()

            assert getattr(parsed_args, 'csv_prefix', False)
            assert getattr(parsed_args, 'csv_interval', None) == 20
            assert getattr(parsed_args, 'csv_flush_interval', None) == 60

            sys.argv = ['grizzly-cli', 'local', 'run', '--csv-prefix', 'static csv prefix', 'test.feature']
            mocker.patch('grizzly_cli.__main__.which', side_effect=['behave'])

            parsed_args = _parse_arguments()

            assert getattr(parsed_args, 'csv_prefix', None) == 'static csv prefix'
            assert getattr(parsed_args, 'csv_interval', None) is None
            assert getattr(parsed_args, 'csv_flush_interval', None) is None
            # // csv logging

            # -T/--testdata-variable
            sys.argv = ['grizzly-cli', 'local', 'run', '-T', 'variable', 'test.feature']
            mocker.patch('grizzly_cli.__main__.which', side_effect=['behave'])

            with pytest.raises(SystemExit) as se:
                _parse_arguments()
            assert se.value.code == 2

            capture = capsys.readouterr()
            assert capture.out == ''
            assert capture.err == 'grizzly-cli: error: -T/--testdata-variable needs to be in the format NAME=VALUE\n'

            sys.argv = ['grizzly-cli', 'local', 'run', '-T', 'key=value', 'test.feature']
            mocker.patch('grizzly_cli.__main__.which', side_effect=['behave'])

            assert environ.get('TESTDATA_VARIABLE_key', None) is None  # noqa: SIM112

            arguments = _parse_arguments()
            assert arguments.command == 'local'
            assert arguments.subcommand == 'run'
            assert arguments.file == 'test.feature'

            assert environ.get('TESTDATA_VARIABLE_key', None) == 'value'  # noqa: SIM112
            # // -T/--testdata-variable
    finally:
        rm_rf(test_context)


def test__parse_argument_dist(capsys: CaptureFixture, mocker: MockerFixture, tmp_path_factory: TempPathFactory) -> None:  # noqa: PLR0915
    test_context = tmp_path_factory.mktemp('test_context')
    (test_context / 'test.feature').write_text('Feature:')

    import sys

    try:
        mocker.patch('grizzly_cli.EXECUTION_CONTEXT', test_context.as_posix())
        with cwd(test_context):
            sys.argv = ['grizzly-cli', 'dist', 'run', 'test.feature']

            mocker.patch('grizzly_cli.__main__.get_distributed_system', side_effect=[None])

            with pytest.raises(SystemExit) as se:
                _parse_arguments()
            assert se.value.code == 2

            capture = capsys.readouterr()
            assert capture.out == ''
            assert capture.err == 'grizzly-cli: error: cannot run distributed\n'

            mocker.patch('grizzly_cli.EXECUTION_CONTEXT', Path.cwd())
            mocker.patch('grizzly_cli.__main__.get_distributed_system', side_effect=['docker'])
            mocker.patch('grizzly_cli.distributed.do_build', side_effect=[0, 4, 0])

            sys.argv = ['grizzly-cli', 'dist', '--limit-nofile', '100', '--registry', 'ghcr.io/biometria-se', 'run', 'test.feature']
            (test_context / 'requirements.txt').write_text('grizzly-loadtester')
            mocker.patch('grizzly_cli.__main__.get_distributed_system', side_effect=['docker'])
            ask_yes_no = mocker.patch('grizzly_cli.__main__.ask_yes_no', autospec=True)

            arguments = _parse_arguments()
            capture = capsys.readouterr()
            assert arguments.limit_nofile == 100
            assert not arguments.yes
            assert arguments.registry == 'ghcr.io/biometria-se/'
            assert capture.out == '!! this will cause warning messages from locust later on\n'
            assert capture.err == ''
            assert ask_yes_no.call_count == 1
            args, _ = ask_yes_no.call_args_list[-1]
            assert args[0] == 'are you sure you know what you are doing?'

            sys.argv = ['grizzly-cli', 'dist', 'run', '--csv-flush-interval', '60', 'test.feature']
            mocker.patch('grizzly_cli.__main__.get_distributed_system', side_effect=['docker'])

            with pytest.raises(SystemExit) as se:
                _parse_arguments()
            assert se.value.code == 2

            capture = capsys.readouterr()
            assert capture.out == ''
            assert capture.err == 'grizzly-cli: error: --csv-flush-interval can only be used in combination with --csv-prefix\n'

            sys.argv = ['grizzly-cli', 'dist', 'build']
            mocker.patch('grizzly_cli.__main__.get_distributed_system', side_effect=['docker'] * 3)
            arguments = _parse_arguments()

            assert not arguments.no_cache
            assert not arguments.force_build
            assert arguments.build
            assert arguments.registry is None

            sys.argv = ['grizzly-cli', 'dist', 'build', '--no-cache', '--registry', 'registry.example.com/biometria-se']
            arguments = _parse_arguments()

            assert arguments.no_cache
            assert arguments.force_build
            assert not arguments.build
            assert arguments.registry == 'registry.example.com/biometria-se/'
    finally:
        rm_rf(test_context)


def test__parse_argument(capsys: CaptureFixture, mocker: MockerFixture, tmp_path_factory: TempPathFactory) -> None:
    test_context = tmp_path_factory.mktemp('test_context')
    (test_context / 'test.feature').write_text('Feature:')

    import sys

    try:
        mocker.patch('grizzly_cli.EXECUTION_CONTEXT', test_context.as_posix())
        with cwd(test_context):
            sys.argv = ['grizzly-cli']

            with pytest.raises(SystemExit) as se:
                _parse_arguments()

            assert se.value.code == 2
            capture = capsys.readouterr()
            assert capture.out == ''
            assert 'usage: grizzly-cli' in capture.err
            assert 'grizzly-cli: error: no command specified' in capture.err

            sys.argv = ['grizzly-cli', 'init', 'test-project']
            arguments = _parse_arguments()

            assert arguments.project == 'test-project'
            assert getattr(arguments, 'subcommand', None) is None
    finally:
        rm_rf(test_context)


def test__inject_additional_arguments_from_metadata(tmp_path_factory: TempPathFactory, capsys: CaptureFixture, mocker: MockerFixture) -> None:
    test_context = tmp_path_factory.mktemp('test_context')

    test_feature_file = test_context / 'test.feature'
    test_feature_file.touch()
    mocker.patch('grizzly_cli.__main__.get_distributed_system', return_value='docker')

    try:
        with cwd(test_context):
            test_feature_file.write_text('# grizzly-cli run --verbose\nFeature:\n')
            sys.argv = ['grizzly-cli', 'dist', 'run', 'test.feature']

            orig_args = _parse_arguments()
            assert not orig_args.verbose

            args = _inject_additional_arguments_from_metadata(orig_args)
            capture = capsys.readouterr()

            assert capture.err == capture.out == ''
            assert args.verbose

            test_feature_file.write_text('# grizzly-cli local --hello\nFeature:\n')
            sys.argv = ['grizzly-cli', 'dist', 'run', 'test.feature']

            orig_args = _parse_arguments()
            args = _inject_additional_arguments_from_metadata(orig_args)

            capture = capsys.readouterr()
            assert capture.err == ''
            assert capture.out == '?? ignoring local --hello\n'

            test_feature_file.write_text('Feature:\n')
            sys.argv = ['grizzly-cli', 'dist', 'run', 'test.feature']

            orig_args = _parse_arguments()
            args = _inject_additional_arguments_from_metadata(orig_args)

            assert args is orig_args

            capture = capsys.readouterr()
            assert capture.err == capture.out == ''

            test_feature_file.write_text('# grizzly-cli --health-timeout 100\nFeature:\n')
            sys.argv = ['grizzly-cli', 'dist', 'run', 'test.feature']

            orig_args = _parse_arguments()
            args = _inject_additional_arguments_from_metadata(orig_args)

            assert args.health_timeout == orig_args.health_timeout

            capture = capsys.readouterr()
            assert capture.err == ''
            assert capture.out == '?? ignoring --health-timeout 100\n'

            test_feature_file.write_text('# grizzly-cli dist --health-timeout 100 --health-retries 101\nFeature:\n# grizzly-cli dist --health-interval 5\n')
            sys.argv = ['grizzly-cli', 'dist', 'run', 'test.feature']

            orig_args = _parse_arguments()
            args = _inject_additional_arguments_from_metadata(orig_args)

            assert args.health_timeout == 100
            assert args.health_retries == 101
            assert args.health_interval == 5

            capture = capsys.readouterr()
            assert capture.err == capture.out == ''
    finally:
        rm_rf(test_context)


def test_main(mocker: MockerFixture, capsys: CaptureFixture) -> None:
    local_mock = mocker.patch('grizzly_cli.__main__.local', side_effect=[0])
    dist_mock = mocker.patch('grizzly_cli.__main__.distributed', side_effect=[1337, 1373])
    init_mock = mocker.patch('grizzly_cli.__main__.init', side_effect=[7331])
    inject_additional_arguments_from_metadata_mock = mocker.patch(
        'grizzly_cli.__main__._inject_additional_arguments_from_metadata',
        return_value=Namespace(command='dist', file='test.feature'),
    )
    mocker.patch(
        'grizzly_cli.__main__._parse_arguments',
        side_effect=[
            Namespace(command='local'),
            Namespace(command='dist'),
            Namespace(command='init'),
            Namespace(command='foobar'),
            KeyboardInterrupt,
            ValueError('hello there'),
            Namespace(command='dist', file='test.feature'),
        ],
    )

    assert main() == 0
    local_mock.assert_called_once_with(SOME(Namespace, command='local'))
    local_mock.reset_mock()
    dist_mock.assert_not_called()
    init_mock.assert_not_called()
    inject_additional_arguments_from_metadata_mock.assert_not_called()

    assert main() == 1337
    local_mock.assert_not_called()
    dist_mock.assert_called_once_with(SOME(Namespace, command='dist'))
    dist_mock.reset_mock()
    init_mock.assert_not_called()
    inject_additional_arguments_from_metadata_mock.assert_not_called()

    assert main() == 7331
    local_mock.assert_not_called()
    dist_mock.assert_not_called()
    init_mock.assert_called_once_with(SOME(Namespace, command='init'))
    init_mock.reset_mock()
    inject_additional_arguments_from_metadata_mock.assert_not_called()

    assert main() == 1

    capture = capsys.readouterr()
    assert capture.err == ''
    assert capture.out == '\nunknown command foobar\n\n!! aborted grizzly-cli\n'

    assert main() == 1

    capture = capsys.readouterr()
    assert capture.err == ''
    assert capture.out == '\n\n!! aborted grizzly-cli\n'

    assert main() == 1

    capture = capsys.readouterr()
    assert capture.err == ''
    assert capture.out == '\nhello there\n\n!! aborted grizzly-cli\n'

    assert main() == 1373
    capture = capsys.readouterr()
    local_mock.assert_not_called()
    dist_mock.assert_called_once_with(SOME(Namespace, command='dist', file='test.feature'))
    dist_mock.reset_mock()
    init_mock.assert_not_called()
    inject_additional_arguments_from_metadata_mock.assert_called_once_with(SOME(Namespace, command='dist', file='test.feature'))



"""Tests for grizzly-cli init."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest
from packaging.version import Version

from test_cli.helpers import rm_rf, run_command

if TYPE_CHECKING:
    from _pytest.tmpdir import TempPathFactory


@pytest.mark.parametrize(
    ('arguments', 'expected'),
    [
        (
            [],
            {'mq_output': 'without IBM MQ support', 'grizzly_output': 'latest grizzly version', 'grizzly_requirements': 'grizzly-loadtester'},
        ),
        (
            ['--grizzly-version', '2.2.2'],
            {'mq_output': 'without IBM MQ support', 'grizzly_output': 'pinned to grizzly version 2.2.2', 'grizzly_requirements': 'grizzly-loadtester==2.2.2'},
        ),
        (
            ['--with-mq'],
            {'mq_output': 'with IBM MQ support', 'grizzly_output': 'latest grizzly version', 'grizzly_requirements': 'grizzly-loadtester[mq]'},
        ),
        (
            ['--with-mq', '--grizzly-version', '1.1.1'],
            {'mq_output': 'with IBM MQ support', 'grizzly_output': 'pinned to grizzly version 1.1.1', 'grizzly_requirements': 'grizzly-loadtester[mq]==1.1.1'},
        ),
    ],
)
def test_e2e_init(arguments: list[str], expected: dict[str, str], tmp_path_factory: TempPathFactory) -> None:
    test_context = tmp_path_factory.mktemp('test_context')

    grizzly_version: Version | None = None
    try:
        grizzly_version_index = arguments.index('--grizzly-version') + 1
        grizzly_version = Version(arguments[grizzly_version_index])
    except ValueError:
        grizzly_version = None

    grizzly_behave_module = 'behave' if grizzly_version is None or grizzly_version >= Version('2.6.0') else 'environment'

    try:
        rc, output = run_command(
            ['grizzly-cli', 'init', 'foobar', '--yes', *arguments],
            cwd=test_context,
        )
        try:
            assert rc == 0
        except AssertionError:
            print(''.join(output))
            raise
        assert (
            ''.join(output)
            == f"""the following structure will be created:

    foobar
    ├── environments
    │   └── foobar.yaml
    ├── features
    │   ├── environment.py
    │   ├── steps
    │   │   └── steps.py
    │   ├── foobar.feature
    │   └── requests
    └── requirements.txt

successfully created project "foobar", with the following options:
  • {expected['mq_output']}
  • {expected['grizzly_output']}
"""
        )

        assert (test_context / 'foobar').is_dir()
        assert (test_context / 'foobar' / 'environments').is_dir()
        requirements_file = test_context / 'foobar' / 'requirements.txt'
        assert requirements_file.is_file()
        assert requirements_file.read_text() == f'{expected["grizzly_requirements"]}\n'

        environments_file = test_context / 'foobar' / 'environments' / 'foobar.yaml'
        assert environments_file.is_file()
        assert (
            environments_file.read_text()
            == """configuration:
  template:
    host: https://localhost
"""
        )

        features_dir = test_context / 'foobar' / 'features'
        assert features_dir.is_dir()

        assert (features_dir / 'requests').is_dir()
        assert list((features_dir / 'requests').rglob('**/*')) == []

        assert (features_dir / 'steps').is_dir()
        steps_file = features_dir / 'steps' / 'steps.py'
        assert steps_file.is_file()
        assert steps_file.read_text() == 'from grizzly.steps import *\n\n'

        environment_file = features_dir / 'environment.py'
        assert environment_file.is_file()
        assert environment_file.read_text() == f'from grizzly.{grizzly_behave_module} import *\n\n'

        feature_file = features_dir / 'foobar.feature'
        assert feature_file.is_file()
        assert (
            feature_file.read_text()
            == """Feature: Template feature file
  Scenario: Template scenario
    Given a user of type "RestApi" with weight "1" load testing "$conf::template.host"
"""
        )
    finally:
        rm_rf(test_context)



"""Grizzly wrapper around argparse."""

from __future__ import annotations

import re
import sys
from argparse import ArgumentParser as CoreArgumentParser
from argparse import Namespace, _SubParsersAction
from typing import TYPE_CHECKING, Any, cast

from grizzly_cli.argparse.bashcompletion import BashCompletionAction
from grizzly_cli.argparse.bashcompletion import hook as bashcompletion_hook
from grizzly_cli.argparse.markdown import MarkdownFormatter, MarkdownHelpAction

if TYPE_CHECKING:  # pragma: no cover
    from collections.abc import Sequence
    from typing import IO

    from _typeshed import SupportsWrite

ArgumentSubParser = _SubParsersAction


class ArgumentParser(CoreArgumentParser):
    def __init__(self, *args: Any, markdown_help: bool = False, bash_completion: bool = False, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

        self.markdown_help = markdown_help
        self.bash_completion = bash_completion

        if self.markdown_help:
            self.add_argument('--md-help', action=MarkdownHelpAction)

        if self.bash_completion:
            self.add_argument('--bash-completion', action=BashCompletionAction)

        self._optionals.title = 'optional arguments'

    def error_no_help(self, message: str) -> None:
        sys.stderr.write(f'{self.prog}: error: {message}\n')
        sys.exit(2)

    def print_help(self, file: SupportsWrite[str] | None = None) -> None:
        """Make help more command line friendly, if there is markdown markers in the text."""
        file = cast('IO[str] | None', file)
        if not self.markdown_help:
            super().print_help(file)
            return

        if self.formatter_class is not MarkdownFormatter:
            original_description = self.description
            original_actions = self._actions

            # code block "markers" are not really nice to have in cli help
            if self.description is not None:
                self.description = '\n'.join([line for line in self.description.split('\n') if '```' not in line])
                self.description = self.description.replace('\n\n', '\n')

            for action in self._actions:
                if action.help is not None:
                    # remove any markdown link markers
                    action.help = re.sub(r'\[([^\]]*)\][\(\[][^\)]*[\)\]]', r'\1', action.help)

        super().print_help(file)

        if self.formatter_class is not MarkdownFormatter:
            self.description = original_description
            self._actions = original_actions

    def parse_args(self, args: Sequence[str] | None = None, namespace: Namespace | None = None) -> Namespace:  # type: ignore[override]
        """Add `--bash-complete` to all parsers, if enabled for parser."""
        if self.bash_completion:
            bashcompletion_hook(self)

        return super().parse_args(args, namespace)



"""Types for grizzly argparse bashcompletion logic."""

from __future__ import annotations

import sys
from argparse import ArgumentTypeError
from fnmatch import filter as fnmatch_filter
from os.path import sep as path_sep
from pathlib import Path
from typing import cast

__all__ = [
    'BashCompletionTypes',
]

ESCAPE_CHARACTERS: dict[str, str | int | None] = {
    ' ': '\\ ',
    '(': '\\(',
    ')': '\\)',
}


class BashCompletionTypes:
    class File:
        def __init__(self, *args: str, missing_ok: bool = False) -> None:
            self.patterns = list(args)
            self.cwd = Path.cwd()
            self.missing_ok = missing_ok

        def __call__(self, value: str) -> str:
            if self.missing_ok:
                return value

            file = Path(value)

            if not file.exists():
                message = f'{value} does not exist'
                raise ArgumentTypeError(message)

            if not file.is_file():
                message = f'{value} is not a file'
                raise ArgumentTypeError(message)

            matches = [match for pattern in self.patterns for match in fnmatch_filter([value], pattern)]

            if len(matches) < 1:
                message = f'{value} does not match {", ".join(self.patterns)}'
                raise ArgumentTypeError(message)

            return value

        @classmethod
        def _transform_path(cls, value: str) -> str:
            value = value.translate(str.maketrans(ESCAPE_CHARACTERS))
            if sys.platform == 'win32':
                value = value.replace('/', path_sep)

            return value

        def list_files(self, value: str | None) -> dict[str, str]:
            matches: dict[str, str] = {}

            if value is not None:
                if sys.platform == 'win32':
                    value = value.replace(path_sep, '/')  # posix style

                for chr_with, chr_replace in ESCAPE_CHARACTERS.items():
                    value = value.replace(cast('str', chr_replace), chr_with)

            for pattern in self.patterns:
                for path in self.cwd.rglob(f'**/{pattern}'):
                    try:
                        path_match = path.relative_to(self.cwd)
                    except ValueError:
                        path_match = path

                    path_match_value = path_match.as_posix()

                    # skip any paths where any part is hidden, or any path that is not (partially) relative to provided value
                    if any(part.startswith('.') for part in path_match.parts) or (value is not None and not path_match_value.startswith(value)):
                        continue

                    match: dict[str, str] | None = None

                    # all paths are treated in posix style
                    if '/' in path_match_value:  # there is a directory in the match
                        try:
                            """
                            find first part that matches with provided value;
                            value = `hel`
                            path_match_value = `hello/example.txt`
                            should be `hello`, and a dir(ectory)
                            """
                            index_match = len(value or '')
                            index_sep = path_match_value.index('/', index_match)
                            match = {self._transform_path(path_match_value[:index_sep]): 'dir'}
                        except ValueError:
                            # no match against provided value, so assume file
                            pass

                    if match is None:
                        match = {self._transform_path(path_match_value): 'file'}

                    matches.update(match)

            return matches



"""Functionality for `grizzly-cli ... run ...`."""

from __future__ import annotations

import os
import sys
from contextlib import suppress
from datetime import datetime
from pathlib import Path
from platform import node as get_hostname
from typing import (
    TYPE_CHECKING,
    TextIO,
    cast,
)

from jinja2 import Environment

import grizzly_cli
from grizzly_cli.argparse.bashcompletion import BashCompletionTypes
from grizzly_cli.utils import (
    ask_yes_no,
    distribution_of_users_per_scenario,
    find_metadata_notices,
    find_variable_names_in_questions,
    get_input,
    logger,
    parse_feature_file,
    requirements,
    rm_rf,
)
from grizzly_cli.utils.configuration import ScenarioTag, get_context_root, load_configuration

if TYPE_CHECKING:  # pragma: no cover
    from argparse import Namespace as Arguments
    from collections.abc import Callable

    from grizzly_cli.argparse import ArgumentSubParser


def create_parser(sub_parser: ArgumentSubParser, parent: str) -> None:
    # grizzly-cli ... run ...
    run_parser = sub_parser.add_parser('run', description='execute load test scenarios specified in a feature file.')
    run_parser.add_argument(
        '--verbose',
        action='store_true',
        required=False,
        help=(
            'changes the log level to `DEBUG`, regardless of what it says in the feature file. gives more verbose logging '
            'that can be useful when troubleshooting a problem with a scenario.'
        ),
    )
    run_parser.add_argument(
        '-T',
        '--testdata-variable',
        action='append',
        type=str,
        required=False,
        help=('specified in the format `<name>=<value>`. avoids being asked for an initial value for a scenario variable.'),
    )
    run_parser.add_argument(
        '-y',
        '--yes',
        action='store_true',
        default=False,
        required=False,
        help='answer yes on any questions that would require confirmation',
    )
    run_parser.add_argument(
        '-e',
        '--environment-file',
        type=BashCompletionTypes.File('*.yaml', '*.yml'),
        required=False,
        default=None,
        help='configuration file with [environment specific information][framework.usage.variables.environment-configuration]',
    )
    run_parser.add_argument(
        '--csv-prefix',
        nargs='?',
        const=True,
        default=None,
        help='write log statistics to CSV files with specified prefix, if no value is specified the description of the gherkin Feature tag will be used, suffixed with timestamp',
    )
    run_parser.add_argument(
        '--csv-interval',
        type=int,
        default=None,
        required=False,
        help='interval that statistics is collected for CSV files, can only be used in combination with `--csv-prefix`',
    )
    run_parser.add_argument(
        '--csv-flush-interval',
        type=int,
        default=None,
        required=False,
        help='interval that CSV statistics is flushed to disk, can only be used in combination with `--csv-prefix`',
    )
    run_parser.add_argument(
        '-l',
        '--log-file',
        type=str,
        default=None,
        required=False,
        help='save all `grizzly-cli` run output in specified log file',
    )
    run_parser.add_argument(
        '--log-dir',
        type=str,
        default=None,
        required=False,
        help='log directory suffix (relative to `requests/logs`) to save log files generated in a scenario',
    )
    run_parser.add_argument(
        '--dump',
        nargs='?',
        default=None,
        const=True,
        help=(
            'Dump parsed contents of file, can be useful when including scenarios from other feature files. If no argument is specified it '
            'will be dumped to stdout, the argument is treated as a filename'
        ),
    )
    run_parser.add_argument(
        '--dry-run',
        action='store_true',
        required=False,
        help='Will setup and run anything up until when locust should start. Useful for debugging feature files when developing new tests',
    )
    run_parser.add_argument(
        '--profile',
        action='store_true',
        required=False,
        help='Enable profiling of grizzly execution, generates a .hprof file upon completion',
    )
    run_parser.add_argument(
        'file',
        nargs='+',
        type=BashCompletionTypes.File('*.feature'),
        help='path to feature file with one or more scenarios',
    )

    if run_parser.prog != f'grizzly-cli {parent} run':  # pragma: no cover
        run_parser.prog = f'grizzly-cli {parent} run'


def should_prompt_questions(args: Arguments, environ: dict) -> None:
    variables = find_variable_names_in_questions(args.file)
    questions = len(variables)
    manual_input = False

    if questions > 0 and not getattr(args, 'validate_config', False):
        logger.info(f'feature file requires values for {questions} variables')

        for variable in variables:
            name = f'TESTDATA_VARIABLE_{variable}'
            value = os.environ.get(name, '')
            while len(value) < 1:
                value = get_input(f'initial value for "{variable}": ')
                manual_input = True

            environ[name] = value

        logger.info('the following values was provided:')
        for key, value in environ.items():
            if not key.startswith('TESTDATA_VARIABLE_'):
                continue
            logger.info(f'{key.replace("TESTDATA_VARIABLE_", "")} = {value}')

        if manual_input:
            ask_yes_no('continue?')


def should_prompt_notices(args: Arguments) -> None:
    notices = find_metadata_notices(args.file)

    if len(notices) > 0:
        output_func = cast('Callable[[str], None]', logger.info) if args.yes else ask_yes_no

        for notice in notices:
            output_func(notice)


def update_grizzly_environment(args: Arguments, environ: dict) -> None:
    if args.environment_file is not None:
        environment_lock_file = load_configuration(Path(args.environment_file).resolve())
        environ.update({'GRIZZLY_CONFIGURATION_FILE': environment_lock_file.as_posix()})

    if args.dry_run:
        environ.update({'GRIZZLY_DRY_RUN': 'true'})

    if args.log_dir is not None:
        environ.update({'GRIZZLY_LOG_DIR': args.log_dir})

    if args.profile:
        environ.update({'GRIZZLY_PROFILE': 'true'})


def build_run_arguments(args: Arguments) -> dict[str, list[str]]:
    run_arguments: dict[str, list[str]] = {
        'master': [],
        'worker': [],
        'common': [],
    }

    if args.verbose:
        run_arguments['common'] += ['--verbose', '--no-logcapture', '--no-capture', '--no-capture-stderr']

    if args.csv_prefix is not None:
        if args.csv_prefix is True:
            parse_feature_file(args.file)
            if grizzly_cli.FEATURE_DESCRIPTION is None:
                message = 'feature file does not seem to have a `Feature:` description to use as --csv-prefix'
                raise ValueError(message)

            csv_prefix = grizzly_cli.FEATURE_DESCRIPTION.replace(' ', '_')
            timestamp = datetime.now().astimezone().strftime('%Y%m%dT%H%M%S')
            args.csv_prefix = f'{csv_prefix}_{timestamp}'

        run_arguments['common'] += [f'-Dcsv-prefix="{args.csv_prefix}"']

        if args.csv_interval is not None:
            run_arguments['common'] += [f'-Dcsv-interval={args.csv_interval}']

        if args.csv_flush_interval is not None:
            run_arguments['common'] += [f'-Dcsv-flush-interval={args.csv_flush_interval}']

    return run_arguments


@requirements(grizzly_cli.EXECUTION_CONTEXT)
def run(args: Arguments, run_func: Callable[[Arguments, dict, dict[str, list[str]]], int]) -> int:
    # always set hostname of host where grizzly-cli was executed, could be useful
    environ: dict = {
        'GRIZZLY_CLI_HOST': get_hostname(),
        'GRIZZLY_EXECUTION_CONTEXT': grizzly_cli.EXECUTION_CONTEXT,
        'GRIZZLY_MOUNT_CONTEXT': grizzly_cli.MOUNT_CONTEXT,
    }

    environment = Environment(autoescape=False, extensions=[ScenarioTag])
    feature_file = Path(args.file)
    environment_lock_file: str | None = None

    # during execution, create a temporary .lock.feature file that will be removed when done
    original_feature_lines = feature_file.read_text().splitlines()
    feature_lock_file = feature_file.parent / f'{feature_file.stem}.lock{feature_file.suffix}'

    try:
        buffer: list[str] = []
        remove_endif = False

        # remove if-statements containing variables (`{$ .. $}`)
        for line in original_feature_lines:
            stripped_line = line.strip()

            if stripped_line[:2] == '{%' and stripped_line[-2:] == '%}':
                if '{$' in stripped_line and '$}' in stripped_line and 'if' in stripped_line:
                    remove_endif = True
                    continue

                if remove_endif and 'endif' in stripped_line:
                    remove_endif = False
                    continue

            buffer.append(line)

        original_feature_content = '\n'.join(buffer)

        template = environment.from_string(original_feature_content)
        environment.extend(feature_file=feature_file, ignore_errors=False)
        feature_content = template.render()
        feature_lock_file.write_text(feature_content)

        if args.dump:
            output: TextIO = Path(args.dump).open('w+') if isinstance(args.dump, str) else sys.stdout  # noqa: SIM115

            try:
                print(feature_content, file=output)
            finally:
                # do not close stdout...
                if output is not sys.stdout:
                    output.close()

            return 0

        args.file = feature_lock_file.as_posix()

        should_prompt_questions(args, environ)
        should_prompt_notices(args)
        update_grizzly_environment(args, environ)

        if not getattr(args, 'validate_config', False):
            distribution_of_users_per_scenario(args, environ)

        run_arguments = build_run_arguments(args)

        return run_func(args, environ, run_arguments)
    finally:
        if environment_lock_file is not None:
            Path(environment_lock_file).unlink(missing_ok=True)

        feature_lock_file.unlink(missing_ok=True)

        with suppress(FileNotFoundError, ValueError):
            tmp_files = get_context_root() / 'files'
            rm_rf(tmp_files)



"""grizzly-cli utils."""

from __future__ import annotations

import logging
import logging.config
import os
import re
import signal as psignal
import stat
import subprocess
import sys
from collections.abc import Callable, Mapping
from contextlib import suppress
from copy import deepcopy
from dataclasses import dataclass, field
from datetime import datetime, timezone
from functools import wraps
from hashlib import sha1
from importlib.util import find_spec
from json import loads as jsonloads
from math import ceil
from os import environ
from pathlib import Path
from shutil import rmtree, which
from tempfile import mkdtemp
from typing import TYPE_CHECKING, Any, ClassVar, Union, cast

import requests
import tomli
from behave.parser import parse_file as feature_file_parser
from jinja2 import Template
from packaging import version as versioning
from progress.spinner import Spinner
from yaml import Dumper

import grizzly_cli

if TYPE_CHECKING:  # pragma: no cover
    from argparse import Namespace as Arguments
    from types import FrameType, TracebackType

    from behave.model import Scenario

logger = logging.getLogger('grizzly-cli')


class SignalHandler:
    handler: Callable[[int, FrameType | None], None]
    signals: dict[int, Union[Callable[[int, FrameType | None], Any], int, None]]

    def __init__(self, handler: Callable[[int, FrameType | None], None], signal: int, *signals: int) -> None:
        self.handler = handler
        self.signals = {signal: None}

        if signals is not None and len(signals) > 0:
            for sig in signals:
                self.signals.update({sig: None})

    def __enter__(self) -> None:
        for signal in self.signals:
            self.signals.update({signal: psignal.getsignal(signal)})
            psignal.signal(signal, self.handler)

    def __exit__(self, exc_type: type[BaseException] | None, exc: BaseException | None, tb: TracebackType | None) -> bool:
        for signal, handler in self.signals.items():
            psignal.signal(signal, handler)

        return exc is None


@dataclass
class RunCommandResult:
    return_code: int
    abort_timestamp: datetime | None = field(init=False, default=None)
    output: list[bytes] | None = field(init=False, default=None)


def run_command(command: list[str], env: dict[str, str] | None = None, *, silent: bool = False, verbose: bool = False, spinner: str | None = None) -> RunCommandResult:
    if env is None:
        env = environ.copy()

    if verbose:
        logger.info('run_command: %s', ' '.join(command))

    process = subprocess.Popen(
        command,
        env=env,
        stderr=subprocess.STDOUT,
        stdout=subprocess.PIPE,
    )

    result = RunCommandResult(return_code=-1)

    if silent:
        result.output = []

    _spinner: Spinner | None = None

    if spinner is not None:  # pragma: no cover
        _spinner = Spinner(f'{spinner} ')

    def sig_handler(*_args: Any, **_kwargs: Any) -> None:  # pragma: no cover
        if result.abort_timestamp is None:
            result.abort_timestamp = datetime.now(timezone.utc)
            process.terminate()

    with SignalHandler(sig_handler, psignal.SIGINT, psignal.SIGTERM):
        try:
            while process.poll() is None:
                stdout = process.stdout
                if stdout is None:
                    break

                if _spinner is not None:  # pragma: no cover
                    _spinner.next()

                output = stdout.readline()
                if not output:
                    break

                if result.output is None:
                    if spinner is None:
                        logger.info(output.decode().rstrip())
                else:
                    result.output.append(output)

            process.terminate()
        except KeyboardInterrupt:
            pass
        finally:
            with suppress(Exception):
                process.kill()

    process.wait()

    if spinner is not None:  # pragma: no cover
        logger.info('')

    result.return_code = process.returncode

    return result


def get_docker_compose_version() -> tuple[int, int, int]:  # pragma: no cover
    output = subprocess.getoutput('docker compose version')  # noqa: S605

    version_line = output.splitlines()[0]

    match = re.match(r'.*version [v]?([1-2]\.[0-9]+\.[0-9]+).*$', version_line)

    return cast('tuple[int, int, int]', tuple([int(part) for part in match.group(1).split('.')])) if match else (0, 0, 0)


def onerror(
    func: Callable,
    path: str,
    exc_info: Union[  # noqa: ARG001
        BaseException,
        tuple[type[BaseException], BaseException, TracebackType | None],
    ],
) -> None:  # pragma: no cover
    """Error handler for shutil.rmtree.

    If the error is due to an access error (read only file)
    it attempts to add write permission and then retries.
    If the error is for another reason it re-raises the error.
    Usage : ``shutil.rmtree(path, onerror=onerror)``
    """
    _path = Path(path)
    # Is the error an access error?
    if not os.access(_path, os.W_OK):
        _path.chmod(stat.S_IWUSR)
        func(path)
    else:
        raise  # noqa: PLE0704


def rm_rf(path: Union[str, Path], *, missing_ok: bool = False) -> None:
    """Remove the path contents recursively, even if some elements
    are read-only.
    """
    p = path.as_posix() if isinstance(path, Path) else path

    try:
        if sys.version_info >= (3, 12):
            rmtree(p, onexc=onerror)
        else:  # pragma: no cover
            rmtree(p, onerror=onerror)
    except FileNotFoundError:  # pragma: no cover
        if not missing_ok:
            raise


def get_dependency_versions(*, local_install: Union[bool, str]) -> tuple[tuple[str | None, list[str] | None], str | None]:  # noqa: C901, PLR0912, PLR0915
    grizzly_requirement: str | None = None
    grizzly_requirement_egg: str
    locust_version: str | None = None
    grizzly_version: str | None = None
    grizzly_extras: list[str] | None = None

    args: tuple[str, ...] = ()
    if isinstance(local_install, str):
        args += (local_install,)
        if not local_install.endswith('requirements.txt'):
            args += ('requirements.txt',)
    else:
        args += ('requirements.txt',)

    project_requirements = Path.joinpath(Path(grizzly_cli.EXECUTION_CONTEXT), *args)

    try:
        with project_requirements.open(encoding='utf-8') as fd:
            for line in fd.readlines():
                if any(pkg in line for pkg in ['grizzly-loadtester', 'grizzly.git'] if not re.match(r'^([\s]+)?#', line)):
                    grizzly_requirement = line.strip()
                    break
    except:
        return (None, None), None

    if grizzly_requirement is None:
        print(f'!! unable to find grizzly dependency in {project_requirements}', file=sys.stderr)
        return ('(unknown)', None), '(unknown)'

    # check if it's a repo or not
    if 'git+' in grizzly_requirement:
        if grizzly_requirement.startswith('git+'):
            url, egg_part = grizzly_requirement.rsplit('#', 1)
            _, grizzly_requirement_egg = egg_part.split('=', 1)
        elif grizzly_requirement.index('@') < grizzly_requirement.index('git+'):
            grizzly_requirement_egg, url = grizzly_requirement.split('@', 1)
            grizzly_requirement_egg = grizzly_requirement_egg.strip()
            url = url.strip()
        else:
            print(f'!! unable to find properly formatted grizzly dependency in {project_requirements}', file=sys.stderr)
            return ('(unknown)', None), '(unknown)'

        url, branch = url.rsplit('@', 1)
        url = url[4:]  # remove git+
        suffix = sha1(grizzly_requirement.encode('utf-8')).hexdigest()  # noqa: S324

        # extras_requirement normalization
        egg = grizzly_requirement_egg.replace('[', '__').replace(']', '__').replace(',', '_')

        tmp_workspace = mkdtemp(prefix='grizzly-cli-')
        repo_destination = Path(tmp_workspace) / f'{egg}_{suffix}'

        try:
            rc = subprocess.check_call(
                [
                    'git',
                    'clone',
                    '--filter=blob:none',
                    '-q',
                    url,
                    repo_destination,
                ],
                shell=False,
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
            )

            if rc != 0:
                print(f'!! unable to clone git repo {url}', file=sys.stderr)
                raise RuntimeError  # abort

            active_branch = branch

            try:
                active_branch = subprocess.check_output(
                    ['git', 'rev-parse', '--abbrev-ref', 'HEAD'],
                    cwd=repo_destination,
                    shell=False,
                    universal_newlines=True,
                ).strip()
                rc = 0
            except subprocess.CalledProcessError as cpe:
                rc = cpe.returncode

            if rc != 0:
                print(f'!! unable to check branch name of HEAD in git repo {url}', file=sys.stderr)
                raise RuntimeError  # abort

            if active_branch != branch:
                try:
                    git_object_type = subprocess.check_output(
                        ['git', 'cat-file', '-t', branch],
                        cwd=repo_destination,
                        shell=False,
                        universal_newlines=True,
                        stderr=subprocess.STDOUT,
                    ).strip()
                except subprocess.CalledProcessError as cpe:  # pragma: no cover
                    if 'Not a valid object name' in cpe.output:
                        git_object_type = 'branch'  # assume remote branch
                    else:
                        print(f'!! unable to determine git object type for {branch}')
                        raise RuntimeError from cpe

                if git_object_type == 'tag':  # pragma: no cover
                    rc += subprocess.check_call(
                        [
                            'git',
                            'checkout',
                            f'tags/{branch}',
                            '-b',
                            branch,
                        ],
                        cwd=repo_destination,
                        shell=False,
                        stdout=subprocess.DEVNULL,
                        stderr=subprocess.DEVNULL,
                    )

                    if rc != 0:  # pragma: no cover
                        print(f'!! unable to checkout tag {branch} from git repo {url}', file=sys.stderr)
                        raise RuntimeError  # abort
                elif git_object_type == 'commit':
                    rc += subprocess.check_call(
                        [
                            'git',
                            'checkout',
                            branch,
                        ],
                        cwd=repo_destination,
                        shell=False,
                        stdout=subprocess.DEVNULL,
                        stderr=subprocess.DEVNULL,
                    )

                    if rc != 0:  # pragma: no cover
                        print(f'!! unable to checkout commit {branch} from git repo {url}', file=sys.stderr)
                        raise RuntimeError  # abort
                else:
                    rc += subprocess.check_call(
                        [
                            'git',
                            'checkout',
                            '-b',
                            branch,
                            '--track',
                            f'origin/{branch}',
                        ],
                        cwd=repo_destination,
                        shell=False,
                        stdout=subprocess.DEVNULL,
                        stderr=subprocess.DEVNULL,
                    )

                    if rc != 0:
                        print(f'!! unable to checkout branch {branch} from git repo {url}', file=sys.stderr)
                        raise RuntimeError  # abort

            if not Path.joinpath(repo_destination, 'pyproject.toml').exists():
                with Path.joinpath(repo_destination, 'grizzly', '__init__.py').open(encoding='utf-8') as fd:
                    version_raw = [line.strip() for line in fd.readlines() if line.strip().startswith('__version__ =')]

                if len(version_raw) != 1:
                    print(f'!! unable to find "__version__" declaration in grizzly/__init__.py from {url}', file=sys.stderr)
                    raise RuntimeError  # abort

                _, grizzly_version, _ = version_raw[-1].split("'")
            else:
                try:
                    with Path.joinpath(repo_destination, 'setup.cfg').open(encoding='utf-8') as fd:
                        version_raw = [line.strip() for line in fd.readlines() if line.strip().startswith('version = ')]

                    if len(version_raw) != 1:
                        print(f'!! unable to find "version" declaration in setup.cfg from {url}', file=sys.stderr)
                        raise RuntimeError  # abort

                    _, grizzly_version = version_raw[-1].split(' = ')
                except FileNotFoundError:
                    if find_spec('setuptools_scm') is None:
                        rc = subprocess.check_call(
                            [
                                sys.executable,
                                '-m',
                                'pip',
                                'install',
                                'setuptools_scm',
                            ]
                        )

                    try:
                        grizzly_version = subprocess.check_output(
                            [
                                sys.executable,
                                '-m',
                                'setuptools_scm',
                            ],
                            shell=False,
                            universal_newlines=True,
                            cwd=repo_destination,
                        ).strip()
                    except subprocess.CalledProcessError as e:  # pragma: no cover
                        print(f'!! unable to get setuptools_scm version from {url}', file=sys.stderr)
                        raise RuntimeError from e  # abort

            try:
                with Path.joinpath(repo_destination, 'requirements.txt').open(encoding='utf-8') as fd:
                    version_raw = [line.strip() for line in fd.readlines() if line.strip().startswith('locust')]

                if len(version_raw) != 1:
                    print(f'!! unable to find "locust" dependency in requirements.txt from {url}', file=sys.stderr)
                    raise RuntimeError  # abort

                match = re.match(r'^locust.{2}(.*?)$', version_raw[-1].strip().split(' ')[0])

                if not match:
                    print(f'!! unable to find locust version in "{version_raw[-1].strip()}" specified in requirements.txt from {url}', file=sys.stderr)
                else:
                    locust_version = match.group(1).strip()
            except FileNotFoundError:  # pragma: no cover
                with Path.joinpath(repo_destination, 'pyproject.toml').open('rb') as fdt:
                    toml_dict = tomli.load(fdt)
                    dependencies = toml_dict.get('project', {}).get('dependencies', [])
                    for dependency in dependencies:
                        if not dependency.startswith('locust'):
                            continue

                        _, locust_version = dependency.strip().split(' ', 1)

                        break
        except RuntimeError:
            pass
        finally:
            rm_rf(tmp_workspace)
    else:
        response = requests.get(
            'https://pypi.org/pypi/grizzly-loadtester/json',
            timeout=10,
        )

        if response.status_code != 200:
            print(f'!! unable to get grizzly package information from {response.url} ({response.status_code})', file=sys.stderr)
        else:
            pypi = jsonloads(response.text)

            grizzly_requirement_egg = grizzly_requirement

            # get grizzly version used in requirements.txt
            if re.match(r'^grizzly-loadtester(\[[^\]]*\])?$', grizzly_requirement):  # latest
                grizzly_version = pypi.get('info', {}).get('version', None)
            else:
                conditions: list[Callable[[versioning.Version], bool]] = []

                match = re.match(r'^(grizzly-loadtester(\[[^\]]*\])?)(.*?)$', grizzly_requirement)

                if match:
                    grizzly_requirement_egg = match.group(1)
                    condition_expression = match.group(3)

                    for condition in condition_expression.split(',', 1):
                        version_string = re.sub(r'^[^0-9]{1,2}', '', condition)
                        condition_version = versioning.parse(version_string)

                        if '>' in condition:
                            compare = condition_version.__le__ if '=' in condition else condition_version.__lt__
                        elif '<' in condition:
                            compare = condition_version.__ge__ if '=' in condition else condition_version.__gt__
                        else:
                            compare = condition_version.__eq__

                        conditions.append(compare)

                matched_version = None

                for available_version in pypi.get('releases', {}):
                    with suppress(versioning.InvalidVersion):
                        version = versioning.parse(available_version)
                        if len(conditions) > 0 and all(compare(version) for compare in conditions):
                            matched_version = version

                if matched_version is None:
                    print(f'!! could not resolve {grizzly_requirement} to one specific version available at pypi', file=sys.stderr)
                else:
                    grizzly_version = str(matched_version)

            if grizzly_version is not None:
                # get version from pypi, to be able to get locust version
                response = requests.get(
                    f'https://pypi.org/pypi/grizzly-loadtester/{grizzly_version}/json',
                    timeout=10,
                )

                if response.status_code != 200:
                    print(f'!! unable to get grizzly {grizzly_version} package information from {response.url} ({response.status_code})', file=sys.stderr)
                else:
                    release_info = jsonloads(response.text)

                    for requires_dist in release_info.get('info', {}).get('requires_dist', []):
                        if not requires_dist.startswith('locust'):
                            continue

                        match = re.match(r'^locust \((.*?)\)$', requires_dist.strip())

                        locust_version = cast('str', match.group(1)) if match else requires_dist.replace('locust', '').strip()

                        if locust_version is not None and locust_version.startswith('=='):
                            locust_version = locust_version[2:]

                        if len(locust_version or '') < 1:
                            print(f'!! unable to find locust version in "{requires_dist.strip()}" specified in pypi for grizzly-loadtester {grizzly_version}', file=sys.stderr)
                            locust_version = '(unknown)'
                        break

                    if locust_version is None:
                        print(f'!! could not find "locust" in requires_dist information for grizzly-loadtester {grizzly_version}', file=sys.stderr)

    if grizzly_version is None:
        grizzly_version = '(unknown)'
    else:
        match = re.match(r'^grizzly-loadtester\[([^\]]*)\]$', grizzly_requirement_egg)

        grizzly_extras = [extra.strip() for extra in match.group(1).split(',')] if match else []

    if locust_version is None:
        locust_version = '(unknown)'

    return (grizzly_version, grizzly_extras), locust_version


def list_images(args: Arguments) -> dict[str, dict[str, str]]:
    images: dict[str, dict[str, str]] = {}
    output = subprocess.check_output(
        [
            f'{args.container_system}',
            'image',
            'ls',
            '--format',
            '{"name": "{{.Repository}}", "tag": "{{.Tag}}", "size": "{{.Size}}", "created": "{{.CreatedAt}}", "id": "{{.ID}}"}',
        ]
    ).decode('utf-8')

    for line in output.split('\n'):
        if len(line) < 1:
            continue
        image = jsonloads(line)
        name = image['name']
        tag = image['tag']
        del image['name']
        del image['tag']

        version = {tag: image}

        if name not in images:
            images[name] = {}

        images[name].update(version)

    return images


def get_default_mtu(args: Arguments) -> str | None:
    try:
        output = subprocess.check_output(
            [
                f'{args.container_system}',
                'network',
                'inspect',
                'bridge',
                '--format',
                '{{ json .Options }}',
            ]
        ).decode('utf-8')

        line, _ = output.split('\n', 1)
        network_options: dict[str, str] = jsonloads(line)

        return network_options.get('com.docker.network.driver.mtu', '1500')
    except:
        return None


def requirements(execution_context: str) -> Callable[[Callable[..., int]], Callable[..., int]]:
    def wrapper(func: Callable[..., int]) -> Callable[..., int]:
        @wraps(func)
        def _wrapper(*args: Any, **kwargs: Any) -> int:
            return func(*args, **kwargs)

        # a bit ugly, but needed for testability
        setattr(func, '__value__', execution_context)  # noqa: B010
        setattr(_wrapper, '__wrapped__', func)  # noqa: B010

        return _wrapper

    return wrapper


def get_distributed_system() -> str | None:
    if which('docker') is not None:
        container_system = 'docker'
    elif which('podman') is not None:
        container_system = 'podman'
        print('!! podman might not work due to buildah missing support for `RUN --mount=type=ssh`: https://github.com/containers/buildah/issues/2835')
    else:
        print('neither "podman" nor "docker" found in PATH')
        return None

    rc, _ = subprocess.getstatusoutput(f'{container_system} compose version')  # noqa: S605

    if rc != 0:
        print(f'"{container_system} compose" not found in PATH')
        return None

    return container_system


def get_input(text: str) -> str:  # pragma: no cover
    return input(text).strip()


def ask_yes_no(question: str) -> None:
    answer = 'undefined'
    while answer.lower() not in ['y', 'n']:
        if answer != 'undefined':
            print('you must answer y (yes) or n (no)')
        answer = get_input(f'{question} [y/n]: ')

        if answer == 'n':
            raise KeyboardInterrupt


def parse_feature_file(file: str) -> None:
    if len(grizzly_cli.SCENARIOS) > 0:
        return

    feature = feature_file_parser(file)

    grizzly_cli.FEATURE_DESCRIPTION = feature.name

    for scenario in feature.scenarios:
        grizzly_cli.SCENARIOS.append(scenario)


def find_metadata_notices(file: str) -> list[str]:
    with Path(file).open('r') as fd:
        return [line.strip().replace('# grizzly-cli:notice ', '') for line in fd if line.strip().startswith('# grizzly-cli:notice ')]


def find_variable_names_in_questions(file: str) -> list[str]:
    unique_variables: set[str] = set()

    parse_feature_file(file)

    for scenario in grizzly_cli.SCENARIOS:
        for step in scenario.steps + scenario.background_steps or []:
            if not step.name.startswith('ask for value of variable'):
                continue

            match = re.match(r'ask for value of variable "([^"]*)"', step.name)

            if not match:
                message = f'could not find variable name in "{step.name}"'
                raise ValueError(message)

            unique_variables.add(match.group(1))

    return sorted(unique_variables)


def _guess_datatype(value: str) -> Union[str, int, float, bool]:
    check_value = value.replace('.', '', 1)

    if check_value[0] == '-':
        check_value = check_value[1:]

    if check_value.isdecimal():
        if float(value) % 1 == 0:
            if value.startswith('0'):
                return str(value)

            return int(float(value))

        return float(value)

    if value.lower() in ['true', 'false']:
        return value.lower() == 'true'

    return value


class ScenarioProperties:
    name: str
    index: int
    identifier: str
    user: str | None
    weight: float
    _iterations: int | None
    _user_count: int | None

    def __init__(
        self,
        name: str,
        index: int,
        weight: float | None = None,
        user: str | None = None,
        iterations: int | None = None,
        user_count: int | None = None,
    ) -> None:
        self.name = name
        self.index = index
        self.user = user
        self._iterations = iterations
        self.weight = weight or 1.0
        self.identifier = f'{index:03}'
        self._user_count = user_count

    @property
    def iterations(self) -> int:
        if self._iterations is None:  # pragma: no cover
            message = 'iterations has not been set'
            raise ValueError(message)

        return self._iterations

    @iterations.setter
    def iterations(self, value: int) -> None:
        self._iterations = value

    @property
    def user_count(self) -> int:
        if self._user_count is None:  # pragma: no cover
            message = 'user count has not been set'
            raise ValueError(message)
        return self._user_count

    @user_count.setter
    def user_count(self, value: int) -> None:
        self._user_count = value

    def is_fulfilled(self) -> bool:
        return self.user is not None and self._iterations is not None and self._user_count is not None


def distribution_of_users_per_scenario(args: Arguments, environ: dict) -> None:  # noqa: C901, PLR0912, PLR0915
    distribution: dict[str, ScenarioProperties] = {}
    variables = {key.replace('TESTDATA_VARIABLE_', ''): _guess_datatype(value) for key, value in environ.items() if key.startswith('TESTDATA_VARIABLE_')}

    def _pre_populate_scenario(scenario: Scenario, index: int) -> None:
        if scenario.name not in distribution:
            distribution[scenario.name] = ScenarioProperties(
                name=scenario.name,
                index=index,
                user=None,
                weight=None,
                iterations=None,
            )

    scenario_user_count_total: int | None = None
    use_weights = True

    for index, scenario in enumerate(grizzly_cli.SCENARIOS):
        scenario_variables: dict = {}
        if len(scenario.steps) < 1:
            message = f'scenario "{scenario.name}" does not have any steps'
            raise ValueError(message)

        _pre_populate_scenario(scenario, index=index + 1)

        if index == 0:  # background_steps is only processed for first scenario in grizzly
            for step in scenario.background_steps or []:
                if (step.name.endswith(' users') or step.name.endswith(' user')) and step.keyword == 'Given':
                    match = re.match(r'"([^"]*)" user(s)?', step.name)
                    if match:
                        scenario_user_count_total = int(round(float(Template(match.group(1)).render(**variables)), 0))

        if scenario_user_count_total is None:
            use_weights = False
            scenario_user_count_total = 0

        for step in (scenario.background_steps or []) + scenario.steps:
            if step.name.startswith('value for variable'):  # pragma: no cover
                match = re.match(r'value for variable "([^"]*)" is "([^"]*)"', step.name)
                if match:
                    try:
                        variable_name = match.group(1)
                        variable_value = Template(match.group(2)).render(**variables, **scenario_variables)
                        scenario_variables.update({variable_name: variable_value})
                    except:  # noqa: S112
                        continue
            elif step.name.startswith('a user of type'):
                match = re.match(r'a user of type "([^"]*)" (with weight "([^"]*)")?.*', step.name)
                if match:
                    distribution[scenario.name].user = match.group(1)
                    distribution[scenario.name].weight = int(float(Template(match.group(3) or '1.0').render(**variables, **scenario_variables)))
            elif step.name.startswith('repeat for'):
                match = re.match(r'repeat for "([^"]*)" iteration[s]?', step.name)
                if match:
                    distribution[scenario.name].iterations = int(round(float(Template(match.group(1)).render(**variables, **scenario_variables)), 0))
            elif any(pattern in step.name for pattern in ['users of type', 'user of type']):
                match = re.match(r'"([^"]*)" user[s]? of type "([^"]*)".*', step.name)
                if match:
                    scenario_user_count = int(round(float(Template(match.group(1)).render(**variables, **scenario_variables)), 0))
                    scenario_user_count_total += scenario_user_count

                    distribution[scenario.name].user_count = scenario_user_count
                    distribution[scenario.name].user = match.group(2)

            if distribution[scenario.name].is_fulfilled():
                break

    scenario_count = len(distribution.keys())
    assert scenario_user_count_total is not None
    if scenario_count > scenario_user_count_total:
        message = f'grizzly needs at least {scenario_count} users to run this feature'
        raise ValueError(message)

    total_weight = 0.0
    total_iterations = 0
    for scenario in distribution.values():
        if scenario.user is None:
            message = f'{scenario.name} does not have a user type'
            raise ValueError(message)

        total_weight += scenario.weight
        total_iterations += scenario.iterations

    if use_weights:
        for scenario in distribution.values():
            scenario.user_count = ceil(scenario_user_count_total * (scenario.weight / total_weight))

        # smooth assigned user count based on weight, so that the sum of scenario.user_count == total_user_count
        total_user_count = sum([scenario.user_count for scenario in distribution.values()])
        user_overflow = total_user_count - scenario_user_count_total

        while user_overflow > 0:
            for scenario in dict(sorted(distribution.items(), key=lambda d: d[1].user_count, reverse=True)).values():
                if scenario.user_count <= 1:
                    continue

                scenario.user_count -= 1
                user_overflow -= 1

                if user_overflow < 1:
                    break

    def print_table_lines(max_length_iterations: int, max_length_users: int, max_length_description: int, max_length_errors: int) -> None:
        line = ['-' * 5, '-|-', '-' * 6, '|-', '-' * max_length_iterations, '|-', '-' * max_length_users, '|-', '-' * max_length_description, '-|']
        if not use_weights:
            line = line[:1] + line[3:]
            line[0] = f'{line[0]}-'

        if max_length_errors > 0:
            line += ['-' * (max_length_errors + 1), '-|']
        logger.info(''.join(line))

    rows: list[str] = []
    max_length_description = len('description')
    max_length_iterations = len('#iter')
    max_length_users = len('#user')
    max_length_errors = len('errors')

    message = f'\nfeature file {args.file} will execute in total {total_iterations} iterations divided on {len(grizzly_cli.SCENARIOS)} scenarios'
    if hasattr(args, 'environment_file') and args.environment_file is not None:
        message = f'{message} with environment file {environ["GRIZZLY_CONFIGURATION_FILE"]}'

    logger.info('%s\n', message)

    errors: dict[str, list[str]] = {}

    for scenario in distribution.values():
        # check for errors
        if scenario.user_count < 1:
            if scenario.name not in errors:
                errors.update({scenario.name: []})

            errors[scenario.name].append('no users assigned')

        if scenario.iterations < 1:
            if scenario.name not in errors:
                errors.update({scenario.name: []})

            errors[scenario.name].append('no iterations')

        # calculate max length out of all rows
        max_length_description = max(len(scenario.name), max_length_description)
        max_length_iterations = max(len(str(scenario.iterations)), max_length_iterations)
        max_length_users = max(len(str(scenario.user_count)), max_length_users)
        max_length_errors = max(len(', '.join(errors.get(scenario.name, []))), max_length_errors)

    # there was no errors, so reset max length for that column, so it won't be included
    if len(errors) < 1:
        max_length_errors = 0

    row_format = ['{:5} ', '  {:>6d}', '  {:>{}}', '  {:>{}}', '  {:<{}}']
    if not use_weights:
        row_format.pop(1)

    if len(errors) > 0:
        row_format.append('   {}')

    for scenario in distribution.values():
        row_format_args: list = [
            scenario.identifier,
            scenario.weight,
            scenario.iterations,
            max_length_iterations,
            scenario.user_count,
            max_length_users,
            scenario.name,
            max_length_description,
        ]

        # remove weights column
        if not use_weights:
            row_format_args.pop(1)

        if len(errors) > 0:
            row_format_args.append(
                ', '.join(errors.get(scenario.name, [])),
            )

        rows.append(''.join(row_format).format(*row_format_args))

    logger.info('each scenario will execute accordingly:\n')
    header_row_args: list = [
        'ident',
        'weight',
        '#iter',
        max_length_iterations,
        '#user',
        max_length_users,
        'description',
        max_length_description,
    ]

    header_row_format = ['{:5} ', '  {:>6}', '  {:>{}}', '  {:>{}}', '  {:<{}}']
    if not use_weights:
        header_row_format.pop(1)
        header_row_args.pop(1)

    if len(errors) > 0:
        header_row_format.append('   {}')
        header_row_args.append('errors')

    logger.info(''.join(header_row_format).format(*header_row_args))
    print_table_lines(max_length_iterations, max_length_users, max_length_description, max_length_errors)
    for row in rows:
        logger.info(row)
    print_table_lines(max_length_iterations, max_length_users, max_length_description, max_length_errors)

    if len(errors) > 0:
        arrow_width = len('ident') + 2 + max_length_iterations + 2 + max_length_users + 2 + max_length_description + 2
        if use_weights:
            arrow_width += len('weight') + 2
        message = f"""{' ' * (1 + arrow_width)}^
+{'-' * arrow_width}+
|
+- there were errors when calculating user distribution and iterations per scenario, adjust user "weight", number of users or iterations per scenario
"""

        raise ValueError(message)

    logger.info('')

    for scenario in distribution.values():
        if scenario.iterations < scenario.user_count:
            message = f'{scenario.name} will have {scenario.user_count} users to run {scenario.iterations} iterations, increase iterations or lower user count'
            raise ValueError(message)

    if not args.yes:
        ask_yes_no('continue?')


def setup_logging(logfile: str | None = None) -> None:
    logging_config: dict = {
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {
            'plain': {
                'format': '%(message)s',
            },
        },
        'handlers': {
            'console': {
                'class': 'logging.StreamHandler',
                'formatter': 'plain',
            },
        },
        'loggers': {
            'grizzly-cli': {
                'handlers': ['console'],
                'level': 'INFO',
                'propagate': False,
            },
        },
        'root': {
            'handlers': ['console'],
            'level': 'INFO',
        },
    }

    if logfile is not None:  # pragma: no cover
        logging_config['handlers']['file'] = {
            'class': 'logging.FileHandler',
            'filename': logfile,
            'formatter': 'plain',
        }

        logging_config['loggers']['grizzly-cli']['handlers'].append('file')
        logging_config['root']['handlers'].append('file')

    logging.config.dictConfig(logging_config)


def unflatten(key: str, value: Any) -> dict:
    paths: list[str] = key.split('.')

    # last node should have the value
    path = paths.pop()
    struct = {path: value}

    # build the struct from the inside out
    paths.reverse()

    for path in paths:
        struct = {path: {**struct}}

    return struct


def flatten(node: dict, parents: list[str] | None = None) -> dict:
    """Flatten a dictionary so each value key is the path down the nested dictionary structure."""
    flat: dict = {}
    if parents is None:
        parents = []

    for key, value in node.items():
        parents.append(key)
        if isinstance(value, dict):
            flat = {**flat, **flatten(value, parents)}
        else:
            flat['.'.join(parents)] = value

        parents.pop()

    return flat


def merge_dicts(merged: dict, source: dict) -> dict:
    """Merge two dicts recursively, where `source` values takes precedance over `merged` values."""
    merged = deepcopy(merged)
    source = deepcopy(source)

    for key in source:
        if key in merged and isinstance(merged[key], dict) and (isinstance(source[key], Mapping) or source[key] is None):
            merged[key] = merge_dicts(merged[key], source[key] or {})
        else:
            value = source[key]
            if isinstance(value, str) and value.lower() == 'none':  # pragma: no cover
                value = None
            merged[key] = value

    return merged


def chunker(value: str, size: int) -> list[str]:
    return [value[x * size : x * size + size] for x in list(range(ceil(len(value) / size)))]


def get_indentation(file: Path) -> int:
    try:
        first_indent_line = file.read_text().splitlines()[1]
        return len(first_indent_line.rstrip()) - len(first_indent_line.strip())
    except IndexError:
        # use 2 as default indentation when it is not possible to detect from file
        return 2


class IndentDumper(Dumper):
    use_indent: ClassVar[int]

    @classmethod
    def use_indentation(cls, target: Path | int) -> type[IndentDumper]:
        cls.use_indent = get_indentation(target) if isinstance(target, Path) else target

        return cls

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

        self.best_indent = self.use_indent

    def increase_indent(self, flow: bool = False, indentless: bool = False) -> None:  # noqa: ARG002, FBT001, FBT002
        return super().increase_indent(flow, False)  # noqa: FBT003



"""Tests for grizzly_cli."""

from __future__ import annotations

from contextlib import suppress
from importlib import reload
from inspect import getfile
from os import environ
from pathlib import Path
from typing import TYPE_CHECKING

from test_cli.helpers import cwd, rm_rf

if TYPE_CHECKING:  # pragma: no cover
    from _pytest.tmpdir import TempPathFactory
    from pytest_mock import MockerFixture


def test___import__(tmp_path_factory: TempPathFactory, mocker: MockerFixture) -> None:
    test_context = tmp_path_factory.mktemp('test_context')

    try:
        with cwd(test_context):
            environ['GRIZZLY_MOUNT_CONTEXT'] = '/srv/grizzly'

            import grizzly_cli

            reload(grizzly_cli)
            mocker.patch.object(grizzly_cli, '__version__', '1.2.3')

            static_context = Path.joinpath(Path(getfile(grizzly_cli)).parent, 'static')

            assert grizzly_cli.__version__ == '1.2.3'
            assert test_context.as_posix() == grizzly_cli.EXECUTION_CONTEXT
            assert grizzly_cli.MOUNT_CONTEXT == '/srv/grizzly'
            assert static_context.as_posix() == grizzly_cli.STATIC_CONTEXT
            assert test_context.name == grizzly_cli.PROJECT_NAME
            assert len(grizzly_cli.SCENARIOS) == 0
    finally:
        rm_rf(test_context)

        with suppress(KeyError):
            del environ['GRIZZLY_MOUNT_CONTEXT']



"""Tests for grizzly_cli.distributed.clean."""

from __future__ import annotations

from argparse import Namespace
from typing import TYPE_CHECKING

from grizzly_cli.distributed.clean import clean
from grizzly_cli.utils import RunCommandResult

if TYPE_CHECKING:
    from pytest_mock import MockerFixture


def test_clean(mocker: MockerFixture) -> None:  # noqa: PLR0915
    import grizzly_cli.distributed.clean

    mocker.patch.object(grizzly_cli.distributed.clean, 'STATIC_CONTEXT', '/srv/grizzly/static-context')
    mocker.patch.object(grizzly_cli.distributed.clean, 'PROJECT_NAME', 'grizzly-cli-test-project')

    arguments = Namespace(networks=True, images=True, project_name='foobar', container_system='docker', id=None)
    mocker.patch('grizzly_cli.distributed.clean.getuser', return_value='root')
    mocker.patch('grizzly_cli.distributed.clean.get_terminal_size', return_value=(1024, 1024))

    run_command_spy = mocker.patch(
        'grizzly_cli.distributed.clean.run_command',
        side_effect=[
            RunCommandResult(return_code=0),
            RunCommandResult(return_code=1),
            RunCommandResult(return_code=2),
            RunCommandResult(return_code=3),
            RunCommandResult(return_code=4),
            RunCommandResult(return_code=5),
            RunCommandResult(return_code=6),
            RunCommandResult(return_code=7),
        ],
    )

    assert clean(arguments) == 0

    assert run_command_spy.call_count == 3

    args, kwargs = run_command_spy.call_args_list[0]
    assert args[0] == [
        'docker',
        'compose',
        '-f',
        '/srv/grizzly/static-context/compose.yaml',
        '-p',
        'foobar-root',
        'rm',
        '-f',
        '-s',
        '-v',
    ]

    actual_env = kwargs.get('env', None)
    assert actual_env is not None

    assert actual_env.get('GRIZZLY_ENVIRONMENT_FILE', None) is not None
    assert actual_env.get('GRIZZLY_PROJECT_NAME', None) == 'foobar'
    assert actual_env.get('GRIZZLY_USER_TAG', None) == 'root'
    assert actual_env.get('GRIZZLY_CONTAINER_TTY', None) == 'false'
    assert actual_env.get('GRIZZLY_LIMIT_NOFILE', None) == '1024'
    assert actual_env.get('COLUMNS', None) == '1024'
    assert actual_env.get('LINES', None) == '1024'

    args, kwargs = run_command_spy.call_args_list[1]
    assert args[0] == [
        'docker',
        'image',
        'rm',
        'foobar:root',
    ]

    assert kwargs == {}

    args, kwargs = run_command_spy.call_args_list[2]
    assert args[0] == [
        'docker',
        'network',
        'rm',
        'foobar-root_default',
    ]

    assert kwargs == {}

    run_command_spy.reset_mock()

    # do not remove images
    arguments.project_name = None
    arguments.id = 'foobar'
    arguments.images = False

    assert clean(arguments) == 3

    assert run_command_spy.call_count == 2

    args, kwargs = run_command_spy.call_args_list[0]
    assert args[0] == [
        'docker',
        'compose',
        '-f',
        '/srv/grizzly/static-context/compose.yaml',
        '-p',
        'grizzly-cli-test-project-foobar-root',
        'rm',
        '-f',
        '-s',
        '-v',
    ]

    assert kwargs.get('env', None) is not None

    args, kwargs = run_command_spy.call_args_list[1]
    assert args[0] == [
        'docker',
        'network',
        'rm',
        'grizzly-cli-test-project-foobar-root_default',
    ]

    assert kwargs == {}

    run_command_spy.reset_mock()

    # do not remove networks
    arguments.images = True
    arguments.networks = False
    arguments.id = None

    assert clean(arguments) == 5

    assert run_command_spy.call_count == 2

    args, kwargs = run_command_spy.call_args_list[0]
    assert args[0] == [
        'docker',
        'compose',
        '-f',
        '/srv/grizzly/static-context/compose.yaml',
        '-p',
        'grizzly-cli-test-project-root',
        'rm',
        '-f',
        '-s',
        '-v',
    ]

    assert kwargs.get('env', None) is not None

    args, kwargs = run_command_spy.call_args_list[1]
    assert args[0] == [
        'docker',
        'image',
        'rm',
        'grizzly-cli-test-project:root',
    ]

    assert kwargs == {}

    run_command_spy.reset_mock()

    # do not remove images or networks
    arguments.images = False

    assert clean(arguments) == 7

    assert run_command_spy.call_count == 1



"""Tests for grizzly_cli.auth."""

from __future__ import annotations

import sys
from contextlib import suppress
from os import environ
from typing import TYPE_CHECKING

import pytest
from grizzly_cli.__main__ import _parse_arguments
from grizzly_cli.auth import auth

if TYPE_CHECKING:  # pragma: no cover
    from _pytest.capture import CaptureFixture
    from _pytest.tmpdir import TempPathFactory
    from pytest_mock import MockerFixture


def test_auth_env(capsys: CaptureFixture, mocker: MockerFixture) -> None:
    try:
        sys.argv = ['grizzly-cli', 'auth']
        arguments = _parse_arguments()

        with pytest.raises(ValueError, match='environment variable OTP_SECRET is not set'):
            auth(arguments)

        capsys.readouterr()

        environ['OTP_SECRET'] = 'f00bar='  # noqa: S105

        with pytest.raises(ValueError, match='unable to generate TOTP code: Non-base32 digit found'):
            auth(arguments)

        environ['OTP_SECRET'] = 'asdfasdf'  # noqa: S105
        mocker.patch('grizzly_cli.auth.TOTP.now', return_value=111111)

        assert auth(arguments) == 0

        capture = capsys.readouterr()

        assert capture.err == ''
        assert capture.out == '111111\n'
    finally:
        with suppress(KeyError):
            del environ['OTP_SECRET']


def test_auth_stdin(capsys: CaptureFixture, mocker: MockerFixture) -> None:
    sys.argv = ['grizzly-cli', 'auth', '-']
    arguments = _parse_arguments()

    mocker.patch('sys.stdin.read', return_value=None)

    with pytest.raises(ValueError, match='OTP secret could not be read from stdin'):
        auth(arguments)

    mocker.patch('sys.stdin.read', return_value=' ')

    with pytest.raises(ValueError, match='OTP secret could not be read from stdin'):
        auth(arguments)

    mocker.patch('sys.stdin.read', return_value='f00bar=')

    with pytest.raises(ValueError, match='unable to generate TOTP code: Non-base32 digit found'):
        auth(arguments)

    capsys.readouterr()

    mocker.patch('grizzly_cli.auth.TOTP.now', return_value=222222)
    mocker.patch('sys.stdin.read', return_value='asdfasdf')

    assert auth(arguments) == 0

    capture = capsys.readouterr()

    assert capture.err == ''
    assert capture.out == '222222\n'


def test_auth_file(capsys: CaptureFixture, mocker: MockerFixture, tmp_path_factory: TempPathFactory) -> None:
    test_context = tmp_path_factory.mktemp('test_context')

    file = test_context / 'secret.txt'

    sys.argv = ['grizzly-cli', 'auth', str(file)]
    arguments = _parse_arguments()

    with pytest.raises(ValueError, match=f'file {file.as_posix()} does not exist'):
        auth(arguments)

    file.write_text(' ')

    with pytest.raises(ValueError, match=f'file {file.as_posix()} does not seem to contain a single line with a valid OTP secret'):
        auth(arguments)

    file.write_text('aasdf\nasdfasdf\n')

    with pytest.raises(ValueError, match=f'file {file.as_posix()} does not seem to contain a single line with a valid OTP secret'):
        auth(arguments)

    file.write_text('hello world\n')

    with pytest.raises(ValueError, match=f'file {file.as_posix()} does not seem to contain a single line with a valid OTP secret'):
        auth(arguments)

    file.write_text('f00bar=\n')

    with pytest.raises(ValueError, match='unable to generate TOTP code: Non-base32 digit found'):
        auth(arguments)

    file.write_text('asdfasdf')
    mocker.patch('grizzly_cli.auth.TOTP.now', return_value=333333)

    assert auth(arguments) == 0

    capture = capsys.readouterr()

    assert capture.err == ''
    assert capture.out == '333333\n'



"""Configuration of pytest."""

from __future__ import annotations

import sys
from os import environ
from pathlib import Path
from typing import TYPE_CHECKING

import pytest

from test_cli.fixtures import (
    End2EndFixture,
)

if TYPE_CHECKING:  # pragma: no cover
    from collections.abc import Generator

    from _pytest.config import Config
    from _pytest.fixtures import SubRequest
    from _pytest.tmpdir import TempPathFactory

E2E_RUN_MODE = environ.get('E2E_RUN_MODE', 'local')
E2E_RUN_DIST = environ.get('E2E_RUN_DIST', 'False').lower() == 'True'.lower()


PYTEST_TIMEOUT = 300 if E2E_RUN_DIST or E2E_RUN_MODE == 'dist' else 150

# programatically load plugin, since pytest is started with `-p no:requests_mock` in pyproject.toml[tool.pytest.ini_options::addopts]
pytest_plugins = ['requests_mock.contrib._pytest_plugin']

if sys.platform == 'darwin' and PYTEST_TIMEOUT > 150:
    PYTEST_TIMEOUT = 500


@pytest.fixture(autouse=True)
def run_before_and_after_tests(tmp_path_factory: TempPathFactory) -> Generator[None, None, None]:
    original_tmp_path = tmp_path_factory._basetemp
    test_root = (Path(__file__).parent / '..' / '..' / '.pytest_tmp').resolve()
    tmp_path_factory._basetemp = test_root
    tmp_path_factory._basetemp.mkdir(exist_ok=True)

    try:
        yield
    finally:
        tmp_path_factory._basetemp = original_tmp_path


# if we're only running E2E tests, set global timeout
def pytest_configure(config: Config) -> None:
    target = getattr(config.known_args_namespace, 'file_or_dir', ['foobar'])
    if len(target) > 0 and 'e2e' in target[0]:
        config._inicache['timeout'] = PYTEST_TIMEOUT


# also, add markers for each test function that starts with test_e2e_, if we're running everything
def pytest_collection_modifyitems(items: list[pytest.Function]) -> None:
    for item in items:
        if item.originalname.startswith('test_e2e_') and item.get_closest_marker('timeout') is None:
            item.add_marker(pytest.mark.timeout(PYTEST_TIMEOUT))


def _e2e_fixture(tmp_path_factory: TempPathFactory, request: SubRequest) -> Generator[End2EndFixture, None, None]:
    distributed = request.param if hasattr(request, 'param') else E2E_RUN_MODE == 'dist'

    with End2EndFixture(tmp_path_factory, distributed=distributed) as fixture:
        yield fixture


e2e_fixture = pytest.fixture(scope='session', params=[False, True] if E2E_RUN_DIST else None)(_e2e_fixture)



"""Functionality for loading grizzly configuration.
- from keyvault
- merging with other configuration files
- ...
"""

from __future__ import annotations

import re
from base64 import b64decode
from contextlib import suppress
from pathlib import Path
from shutil import which
from textwrap import dedent
from typing import TYPE_CHECKING, ClassVar, cast

import yaml
from azure.identity import AzureCliCredential, ChainedTokenCredential, ManagedIdentityCredential
from azure.keyvault.secrets import KeyVaultSecret, SecretClient
from behave.parser import parse_feature
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives._serialization import PBES, KeySerializationEncryption, KeySerializationEncryptionBuilder, PrivateFormat
from cryptography.hazmat.primitives.serialization import pkcs12
from jinja2 import Environment
from jinja2.lexer import Token, TokenStream
from jinja2_simple_tags import StandaloneTag

from grizzly_cli.utils import IndentDumper, logger, merge_dicts, run_command, unflatten

if TYPE_CHECKING:  # pragma: no cover
    from collections.abc import Iterable

    from behave.model import Scenario
    from cryptography.hazmat.primitives.asymmetric.types import PrivateKeyTypes
    from cryptography.x509 import Certificate


def get_context_root() -> Path:
    possible_context_roots = Path.cwd().rglob('environment.py')

    context_root: Path | None = None

    for possible_context_root in possible_context_roots:
        if any(ignore in possible_context_root.as_posix() for ignore in ['.venv', '.env', 'node_modules']):
            continue

        if context_root is None:
            context_root = possible_context_root
            continue

        if possible_context_root.as_posix().count('/') < context_root.as_posix().count('/'):
            context_root = possible_context_root

    if context_root is None:
        message = 'context root not found, are you in a grizzly project?'
        raise ValueError(message)

    return context_root.parent


class ScenarioTag(StandaloneTag):
    tags: ClassVar[set[str]] = {'scenario'}

    def preprocess(
        self,
        source: str,
        name: str | None,
        filename: str | None = None,
    ) -> str:
        self._source = source

        return cast('str', super().preprocess(source, name, filename))

    @classmethod
    def get_scenario_text(cls, name: str, file: Path) -> str:
        content = file.read_text()

        content_skel = re.sub(r'\{%.*%\}', '', content)
        content_skel = re.sub(r'\{\$.*\$\}', '', content_skel)

        assert len(content.splitlines()) == len(content_skel.splitlines()), 'oops, there is not a 1:1 match between lines!'

        feature = parse_feature(content_skel, filename=file.as_posix())
        scenarios = cast('list[Scenario]', feature.scenarios)
        lines = content.splitlines()

        scenario_index: int
        for index, scenario in enumerate(scenarios):
            if scenario.name == name:
                scenario_index = index
                break

        # check if there are scenarios after our scenario in the source
        next_scenario: Scenario | None = None
        with suppress(IndexError):
            next_scenario = scenarios[scenario_index + 1]

        if next_scenario is None:  # last scenario, take everything until the end
            scenario_lines = lines[scenario.line :]
        else:  # take everything up until where the next scenario starts
            scenario_lines = lines[scenario.line : next_scenario.line - 1]
            if scenario_lines[-1] == '':  # if last line is an empty line, lets remove it
                scenario_lines.pop()

        # remove any scenario text/comments
        if scenario_lines[0].strip() == '"""':
            try:
                offset = scenario_lines[1:].index(scenario_lines[0]) + 1 + 1
            except:
                offset = 0

            scenario_lines = scenario_lines[offset:]

        # first line can have incorrect indentation
        scenario_lines[0] = dedent(scenario_lines[0])

        return '\n'.join(scenario_lines)

    def render(self, scenario: str, feature: str, **variables: str) -> str:
        feature_file = Path(feature)

        # check if relative to parent feature file
        if not feature_file.exists():
            feature_file = (self.environment.feature_file.parent / feature).resolve()

        scenario_content = self.get_scenario_text(scenario, feature_file)

        ignore_errors = getattr(self.environment, 'ignore_errors', False)

        # <!-- sub-render included scenario
        errors_unused: set[str] = set()
        errors_undeclared: set[str] = set()

        # tag has specified variables, so lets "render"
        for name, value in variables.items():
            variable_template = f'{{$ {name} $}}'
            if variable_template not in scenario_content:
                errors_unused.add(name)
                continue

            scenario_content = scenario_content.replace(variable_template, str(value))

        # look for sub-variables that has not been rendered
        if not ignore_errors:
            if '{$' in scenario_content and '$}' in scenario_content:
                matches = re.finditer(r'\{\$ ([^$]+) \$\}', scenario_content, re.MULTILINE)

                for match in matches:
                    errors_undeclared.add(match.group(1))

            if len(errors_undeclared) + len(errors_unused) > 0:
                scenario_identifier = f'{feature}#{scenario}'
                buffer_error: list[str] = []
                if len(errors_unused) > 0:
                    errors_unused_message = '\n  '.join(errors_unused)
                    buffer_error.append(f'the following variables has been declared in scenario tag but not used in {scenario_identifier}:\n  {errors_unused_message}')
                    buffer_error.append('')

                if len(errors_undeclared) > 0:
                    errors_undeclared_message = '\n  '.join(errors_undeclared)
                    buffer_error.append(f'the following variables was used in {scenario_identifier} but was not declared in scenario tag:\n  {errors_undeclared_message}')
                    buffer_error.append('')

                message = '\n'.join(buffer_error)
                raise ValueError(message)

        # check if we have nested statements (`{% .. %}`), and render again if that is the case
        if '{%' in scenario_content and '%}' in scenario_content:
            environment = self.environment.overlay()
            environment.feature_file = feature_file
            template = environment.from_string(scenario_content)
            scenario_content = template.render()
        # // -->

        return scenario_content

    def filter_stream(self, stream: TokenStream) -> TokenStream | Iterable[Token]:  # type: ignore[return]  # noqa: PLR0912
        """Everything outside of `{% scenario ... %}` (and `{% if ... %}...{% endif %}`) should be treated as "data", e.g. plain text.

        Overloaded from `StandaloneTag`, must match method signature, which is not `Generator`, even though we yield
        the result instead of returning.
        """
        in_scenario = False
        in_block_comment = False
        in_condition = False
        in_variable = False

        variable_begin_pos = -1
        variable_end_pos = 0
        block_begin_pos = -1
        block_end_pos = 0
        source_lines = self._source.splitlines()

        for token in stream:
            if token.type == 'block_begin':
                if stream.current.value in self.tags:  # {% scenario ... %}
                    in_scenario = True
                    current_line = source_lines[token.lineno - 1].lstrip()
                    in_block_comment = current_line.startswith('#')
                    block_begin_pos = self._source.index(token.value, block_begin_pos + 1)
                elif stream.current.value in ['if', 'endif']:  # {% if <condition> %}, {% endif %}
                    in_condition = True

            if in_scenario:
                if token.type == 'block_end' and in_block_comment:
                    in_block_comment = False
                    block_end_pos = self._source.index(token.value, block_begin_pos)
                    token_value = self._source[block_begin_pos : block_end_pos + len(token.value)]
                    filtered_token = Token(token.lineno, 'data', token_value)
                elif in_block_comment:
                    continue
                else:
                    filtered_token = token
            elif in_condition:
                filtered_token = token
            else:
                if token.type == 'variable_begin':
                    # Find variable start in the source
                    variable_begin_pos = self._source.index(token.value, variable_begin_pos + 1)
                    in_variable = True
                    continue
                elif token.type == 'variable_end':
                    # Find variable end in the source
                    variable_end_pos = self._source.index(token.value, variable_begin_pos)
                    # Extract the variable definition substring and use as token value
                    token_value = self._source[variable_begin_pos : variable_end_pos + len(token.value)]
                    in_variable = False
                elif in_variable:  # Variable templates is yielded when the whole block has been processed
                    continue
                else:
                    token_value = token.value

                filtered_token = Token(token.lineno, 'data', token_value)

            yield filtered_token

            if token.type == 'block_end':
                if in_scenario:
                    in_scenario = False

                if in_condition:
                    in_condition = False


class MergeYamlTag(StandaloneTag):  # pragma: no cover
    tags: ClassVar[set[str]] = {'merge'}

    def preprocess(
        self,
        source: str,
        name: str | None,
        filename: str | None = None,
    ) -> str:
        self._source = source
        return cast('str', super().preprocess(source, name, filename))

    def render(self, filename: str, *filenames: str) -> str:
        buffer: list[str] = []

        files = [filename, *filenames]

        for file in files:
            merge_file = Path(file)

            # check if relative to parent feature file
            if not merge_file.exists():
                merge_file = (self.environment.source_file.parent / merge_file).resolve()

            if not merge_file.exists():
                raise FileNotFoundError(merge_file)

            merge_content = merge_file.read_text()

            if merge_content[0:3] != '---':
                buffer.append('---')

            buffer.append(merge_content)

        if self._source[0:3] != '---':
            buffer.append('---')

        return '\n'.join(buffer)

    def filter_stream(self, stream: TokenStream) -> TokenStream | Iterable[Token]:  # type: ignore[return]
        """Everything outside of `{% merge ... %}` should be treated as "data", e.g. plain text."""
        in_merge = False
        in_variable = False
        in_block_comment = False

        variable_begin_pos = -1
        variable_end_pos = 0
        block_begin_pos = -1
        block_end_pos = 0
        source_lines = self._source.splitlines()

        for token in stream:
            if token.type == 'block_begin' and stream.current.value in self.tags:
                in_merge = True
                current_line = source_lines[token.lineno - 1].lstrip()
                in_block_comment = current_line.startswith('#')
                block_begin_pos = self._source.index(token.value, block_begin_pos + 1)

            if not in_merge:
                if token.type == 'variable_end':
                    # Find variable end in the source
                    variable_end_pos = self._source.index(token.value, variable_begin_pos)
                    # Extract the variable definition substring and use as token value
                    token_value = self._source[variable_begin_pos : variable_end_pos + len(token.value)]
                    in_variable = False
                elif token.type == 'variable_begin':
                    # Find variable start in the source
                    variable_begin_pos = self._source.index(token.value, variable_begin_pos + 1)
                    in_variable = True
                else:
                    token_value = token.value

                if in_variable:
                    # While handling in-variable tokens, withhold values until
                    # the end of the variable is reached
                    continue

                filtered_token = Token(token.lineno, 'data', token_value)
            elif token.type == 'block_end' and in_block_comment:
                in_block_comment = False
                block_end_pos = self._source.index(token.value, block_begin_pos)
                token_value = self._source[block_begin_pos : block_end_pos + len(token.value)]
                filtered_token = Token(token.lineno, 'data', token_value)
            elif in_block_comment:
                continue
            else:
                filtered_token = token

            yield filtered_token

            if in_merge and token.type == 'block_end':
                in_merge = False


def get_keyvault_client(url: str) -> SecretClient:
    credential = ChainedTokenCredential(ManagedIdentityCredential(), AzureCliCredential())

    return SecretClient(vault_url=url, credential=credential)


def _get_metadata(content_type: str, name: str) -> str | None:
    try:
        start = content_type.index(f'{name}:') + len(name) + 1
    except ValueError:
        return None

    try:
        end = content_type.index(',', start)
    except ValueError:
        end = len(content_type)

    try:
        value = content_type[start:end]
    except IndexError:
        value = None

    return value


def _create_safe_file_and_parent(file: Path) -> Path:
    file.parent.mkdir(parents=True, exist_ok=True)
    file.parent.chmod(0o700)
    file.touch()
    file.chmod(0o600)

    return file


def _create_relative_path(root: Path, file: Path, *, no_suffix: bool = False) -> str:
    if no_suffix:
        file = file.with_suffix('')

    return file.as_posix().replace(root.as_posix(), '')[1:]


def _write_mqm_cert(
    root: Path,
    label: str,
    password: str | None,
    private_key: pkcs12.PKCS12PrivateKeyTypes | None,
    public_certificate: Certificate | None,
    additional_certificates: list[Certificate] | None,
    encryption_algorithm: KeySerializationEncryption,
) -> str:
    p12_file = _create_safe_file_and_parent(root / 'files' / f'{label}.p12')
    cms_file = p12_file.parent / f'{label}.kdb'

    if cms_file.exists():
        cms_file.unlink(missing_ok=True)

        for file_ext in ['rdb', 'sth']:
            cms_file.with_suffix(f'.{file_ext}').unlink(missing_ok=True)

    logger.debug('p12 file: %s', p12_file.as_posix())

    p12_data = pkcs12.serialize_key_and_certificates(
        name=label.encode('utf-8'),
        key=private_key,
        cert=public_certificate,
        cas=additional_certificates,
        encryption_algorithm=encryption_algorithm,
    )

    p12_file.write_bytes(p12_data)

    runmqakm_path = which('runmqakm')

    if runmqakm_path is None:
        message = "runmqakm could not be found, install IBM MQC Redist, and make sure that it's bin/ directory is added to PATH"
        raise ValueError(message)

    runmqakm_cmd: list[str] = [
        runmqakm_path,
        '-keydb',
        '-convert',
        '-new_format',
        'cms',
        '-old_format',
        'p12',
        '-db',
        p12_file.as_posix(),
        '-target',
        cms_file.as_posix(),
    ]

    if password is not None:
        runmqakm_cmd += [
            '-pw',
            password,
            '-stash',
        ]

    relative_file = cms_file.as_posix().replace(root.as_posix(), '')[1:]

    try:
        result = run_command(runmqakm_cmd, silent=True)

        if result.return_code != 0:
            for line in result.output or []:
                logger.error(line.decode('utf-8').strip())

            message = f'failed to create {relative_file}'
            raise ValueError(message)
    finally:
        p12_file.unlink()
        cms_file.with_suffix('.crl').unlink(missing_ok=True)

    for file in cms_file.parent.glob(f'{cms_file.stem}.*'):
        relative_cms_file = _create_relative_path(root, file)
        logger.info('wrote %s', relative_cms_file)

    return _create_relative_path(root, cms_file, no_suffix=True)


def _write_pem_private(root: Path, name: str, encryption_algorithm: KeySerializationEncryption, private_key: PrivateKeyTypes) -> str:
    private_key_file = _create_safe_file_and_parent(root / 'files' / f'{name}.key')

    private_key_pem = private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.TraditionalOpenSSL,
        encryption_algorithm=encryption_algorithm,
    )

    private_key_file.write_bytes(private_key_pem)

    return _create_relative_path(root, private_key_file)


def _write_pem_public(root: Path, name: str, public_certificate: Certificate, additional_certificates: list[Certificate]) -> str:
    certificate_file = _create_safe_file_and_parent(root / 'files' / f'{name}.crt')

    certificate_data: list[bytes] = []

    for certificate in [public_certificate, *additional_certificates]:
        certificate_pem = certificate.public_bytes(encoding=serialization.Encoding.PEM)
        certificate_data.append(certificate_pem)

    certificate_file.write_bytes(b''.join(certificate_data))

    return _create_relative_path(root, certificate_file)


def _write_file(root: Path, content_type: str, encoded_content: str) -> str:
    file_name = _get_metadata(content_type, 'file')

    if file_name is None:
        message = 'could not find `file:` in content type'
        raise ValueError(message)

    file = _create_safe_file_and_parent(root / 'files' / file_name)

    complete = True

    if 'chunk' in content_type:  # 0 = chunk:N, 1 = chunks:T
        chunk_index = int(_get_metadata(content_type, 'chunk') or 0)
        chunk_count = int(_get_metadata(content_type, 'chunks') or 1)
        content_buffer: list[str] = []

        logger.debug('%% processing chunk %d of %d for %s', chunk_index + 1, chunk_count, file_name)

        if chunk_index == 0 and file.exists():
            file.unlink()

        try:
            content_buffer.append(file.read_text())
        except FileNotFoundError:
            if chunk_index > 0:  # pragma: no cover
                raise

        content_buffer.append(encoded_content)

        if chunk_index < chunk_count - 1:
            file.write_text(''.join(content_buffer))
            complete = False
        else:
            logger.debug('%% completed %s with chunk %d of %d', file_name, chunk_index + 1, chunk_count)
            encoded_content = ''.join(content_buffer)
            complete = True

    relative_file = _create_relative_path(root, file)

    if complete:
        content = b64decode(encoded_content)
        file.write_bytes(content)
        logger.debug('%% wrote %s (size %d)', relative_file, len(content))

    return relative_file


def _import_files(client: SecretClient, root: Path, secret: KeyVaultSecret) -> str:
    if secret.value is None:
        message = f'secret {secret.name} has no value'
        raise ValueError(message)

    file_keys = secret.value.split(',')

    for file_key in file_keys:
        file_secret = client.get_secret(file_key)
        if file_secret.properties.content_type is None:
            message = f'secret {file_key} has no content type'
            raise ValueError(message)

        if file_secret.value is None:
            message = f'secret {file_key} has no value'
            raise ValueError(message)

        logger.debug('%% keyvault key %s', file_key)

        # nested files
        if file_secret.properties.content_type.startswith('files'):
            conf_value = _import_files(client, root, file_secret)
            logger.debug('%% nested files configuration value is %s', conf_value)
        else:
            conf_value = _write_file(root, file_secret.properties.content_type, file_secret.value)

    return conf_value


def load_configuration(file: Path) -> Path:
    if not file.exists():
        message = f'{file.as_posix()} does not exist'
        raise ValueError(message)

    if file.suffix not in ['.yml', '.yaml']:
        message = 'configuration file must have file extension yml or yaml'
        raise ValueError(message)

    configuration = load_configuration_file(file)

    load_from_keyvault = ((configuration or {}).get('configuration', None) or {}).get('keyvault', None)

    if load_from_keyvault is not None:
        client = get_keyvault_client(load_from_keyvault)
        context_root = get_context_root()
        environment = configuration.get('configuration', {}).get('env', file.stem)

        loaded_keyvault_configuration, number_of_keyvault_secrets = load_configuration_keyvault(client, environment, context_root, filter_keys=None)
        keyvault_configuration = {'configuration': loaded_keyvault_configuration}

        configuration = merge_dicts(keyvault_configuration, configuration)

        logger.info('loaded %d secrets from keyvault %s', number_of_keyvault_secrets, load_from_keyvault)

    environment_lock_file = _create_safe_file_and_parent(file.parent / f'{file.stem}.lock{file.suffix}')

    with environment_lock_file.open('w') as fd:
        yaml.dump(configuration, fd, Dumper=IndentDumper.use_indentation(file), default_flow_style=False, sort_keys=False, allow_unicode=True)

    return file.with_name(f'{file.stem}.lock{file.suffix}')


def load_configuration_file(file: Path) -> dict:
    """Load a grizzly environment file and flatten the structure."""
    configuration: dict = {}

    environment = Environment(autoescape=False, extensions=[MergeYamlTag])
    environment.extend(source_file=file)
    loader = yaml.SafeLoader

    yaml_template = environment.from_string(file.read_text())
    yaml_content = yaml_template.render()

    yaml_configurations = list(yaml.load_all(yaml_content, Loader=loader))
    yaml_configurations.reverse()
    for yaml_configuration in yaml_configurations:
        configuration = merge_dicts(configuration, yaml_configuration)

    logger.debug('configuration: %r', configuration)

    return configuration


def filter_secrets(client: SecretClient, environment_filter: list[str]) -> dict[str, str]:
    secret_properties = client.list_properties_of_secrets()
    keys: dict[str, str] = {}

    # loop through all secrets to find the ones that match the environment filter
    for secret_property in secret_properties:
        if (
            secret_property.name is None
            or not secret_property.name.startswith('grizzly--')
            or 'noconf' in (secret_property.content_type or '')  # skip chunked secrets, we'll get them later
        ):
            continue

        _, target_environment, name = secret_property.name.split('--', 2)

        # hm, chunk suffix but no chunk in name?
        if '--' in name:
            logger.debug(f'secret {secret_property.name} with name {name} contains `--`, skipping')
            continue

        name = name.replace('-', '.')

        if target_environment not in environment_filter:
            continue

        keys.update({secret_property.name: name})

    return keys


def get_certificate_encryption_algorithm(client: SecretClient, password_key: str | None) -> tuple[KeySerializationEncryption, str | None]:
    # build encryption algorithm
    if password_key is not None:
        password_secret = client.get_secret(password_key)
        password = password_secret.value
    else:
        password = None

    if password is not None:
        encryption_algorithm = KeySerializationEncryptionBuilder(
            PrivateFormat.PKCS12,
            _key_cert_algorithm=PBES.PBESv1SHA1And3KeyTripleDESCBC,
        ).build(password.encode('utf-8'))
    else:
        encryption_algorithm = serialization.NoEncryption()

    return encryption_algorithm, password


def encode_certificate(client: SecretClient, root: Path, secret: KeyVaultSecret, content_type: str) -> str:
    assert secret.value is not None
    arguments: dict[str, str] = {}

    for part in secret.value.split(',', 1) + content_type.split(','):
        argument, value = part.split(':', 1)
        arguments.update({argument: value})

    cert_key = arguments['cert']
    cert_secret = client.get_secret(cert_key)

    if cert_secret.value is None:
        message = f'unable to download certificate secret {cert_key}'
        raise ValueError(message)

    if arguments.get('name') is None:
        name, _ = cert_key.split('-', 1)
        arguments.update({'name': name.lower()})

    certificate = b64decode(cert_secret.value)

    private_key, public_certificate, additional_certificates = pkcs12.load_key_and_certificates(data=certificate, password=None)

    encryption_algorithm, password = get_certificate_encryption_algorithm(client, arguments.get('pass'))

    # write files
    cert_format = arguments.get('format')
    if cert_format == 'pem-private':
        if private_key is None:
            message = f'could not find a private key in {cert_key}'
            raise ValueError(message)

        conf_value = _write_pem_private(root, arguments['name'], encryption_algorithm, private_key)
    elif cert_format == 'pem-public':
        if public_certificate is None:
            message = f'could not find a public certificate in {cert_key}'
            raise ValueError(message)

        conf_value = _write_pem_public(root, arguments['name'], public_certificate, additional_certificates)
    elif cert_format == 'mqm':
        conf_value = _write_mqm_cert(
            root,
            arguments['name'],
            password,
            cast('pkcs12.PKCS12PrivateKeyTypes | None', private_key),
            public_certificate,
            additional_certificates,
            encryption_algorithm,
        )
    else:
        message = f'{cert_format} is not a supported certificate format'
        raise ValueError(message)

    return conf_value


def encode_secret_value(client: SecretClient, root: Path, secret: KeyVaultSecret, secret_key: str) -> str:
    assert secret.value is not None
    assert secret.properties.content_type is not None

    content_type = secret.properties.content_type

    if content_type.startswith('files'):
        conf_value = _import_files(client, root, secret)
        if all(keyword in secret_key for keyword in ['mq', 'key']):
            conf_value = Path(conf_value).with_suffix('').as_posix()
    elif content_type.startswith('file:'):
        conf_value = _write_file(root, content_type, secret.value)
    elif content_type.startswith('format:') and secret.value.startswith('cert:'):
        conf_value = encode_certificate(client, root, secret, content_type)
    else:
        message = f'unknown content type for secret {secret_key}: {content_type}'
        raise ValueError(message)

    return conf_value


def load_configuration_keyvault(client: SecretClient, environment: str, root: Path, *, filter_keys: list[str] | None) -> tuple[dict, int]:
    keys = filter_secrets(client, environment_filter=['global', environment])
    configuration: dict = {}
    imported_secrets = 0

    # get the actual value for all secrets that matched environment filter
    for secret_key, conf_key in keys.items():
        secret = client.get_secret(secret_key)

        if filter_keys is not None and not any(conf_key.startswith(filter_key) for filter_key in filter_keys):
            continue

        content_type = secret.properties.content_type

        if secret.value is None:
            logger.error(f'! secret {secret_key} has no value, skipping')
            continue

        conf_value = secret.value

        no_conf = False

        if content_type is not None:
            no_conf = 'noconf' in content_type
            conf_value = encode_secret_value(client, root, secret, secret_key)

        if not no_conf:
            logger.debug('mapping %s to %s', conf_key, conf_value if content_type is not None else '******')
            configuration_branch = unflatten(conf_key, conf_value)
            configuration = merge_dicts(configuration_branch, configuration)
            imported_secrets += 1

    logger.debug('keyvault configuration: %r', configuration)

    return configuration, imported_secrets
